# -*- coding: utf-8 -*-
"""Cursor movement helpers for Vim emulation."""

# Standard Libraries
import ast
import re
from bisect import bisect_left, bisect_right
from typing import TYPE_CHECKING

# Third Party Libraries
from qtpy.QtCore import QPoint, QRegularExpression
from qtpy.QtGui import QTextCursor, QTextDocument
from qtpy.QtWidgets import QTextEdit
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.motion import MotionInfo, MotionType
from spyder_okvim.utils.search_helpers import SearchHelper
from spyder_okvim.utils.text_constants import BRACKET_PAIR

if TYPE_CHECKING:  # pragma: no cover - typing helpers
    from spyder_okvim.utils.cell_helpers import CellRegion

WHITE_SPACE = " \t"


class MotionHelper:
    """Utilities for computing cursor motions."""

    def __init__(self, vim_status):
        """Create helper bound to the given Vim status object.

        Args:
            vim_status: Object providing editor utilities.
        """
        self.vim_status = vim_status
        self.get_editor = vim_status.get_editor
        self.get_cursor = vim_status.get_cursor
        self.set_cursor = vim_status.set_cursor
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos

        self.search_helper = SearchHelper(vim_status, self._set_motion_info)

        self.find_cmd_map = {
            "f": self.find_ch,
            "F": self.rfind_ch,
            "t": self.t,
            "T": self.T,
            "s": self.search_helper.sneak,
            "S": self.search_helper.rsneak,
            "z": self.search_helper.sneak,
            "Z": self.search_helper.rsneak,
        }

    def _set_motion_info(
        self,
        cur_pos: int | None,
        sel_start: int | None = None,
        sel_end: int | None = None,
        motion_type: int = MotionType.LineWise,
    ):
        """Build a :class:`MotionInfo` object.

        Args:
            cur_pos: The resulting cursor position.
            sel_start: Selection start position.
            sel_end: Selection end position.
            motion_type: Type of the motion.

        Returns:
            MotionInfo: Computed motion information.
        """
        return MotionInfo(
            cursor_pos=cur_pos,
            sel_start=sel_start,
            sel_end=sel_end,
            motion_type=motion_type,
        )

    # ------------------------------------------------------------------
    # Cell navigation helpers
    # ------------------------------------------------------------------
    def _locate_cell_index(
        self, cells: list["CellRegion"], cursor_pos: int
    ) -> int | None:
        """Return the index of the cell containing ``cursor_pos``."""
        if not cells:
            return None

        for idx, cell in enumerate(cells):
            if cell.start_position <= cursor_pos < cell.end_position:
                return idx

        if cursor_pos < cells[0].start_position:
            return 0

        for idx in range(len(cells) - 1, -1, -1):
            if cursor_pos >= cells[idx].start_position:
                return idx

        return None

    def prev_cell(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return motion info for the previous code cell."""
        cells = self.vim_status.get_cells()
        if len(cells) <= 1:
            return self._set_motion_info(None)

        cursor = self.get_cursor()
        cursor_pos = cursor.position()
        current_idx = self._locate_cell_index(cells, cursor_pos)
        if current_idx is None:
            return self._set_motion_info(None)

        target_idx = current_idx - num
        if target_idx < 0:
            return self._set_motion_info(None)

        target_cell = cells[target_idx]
        target_pos = target_cell.start_position

        current_cell = cells[current_idx]
        current_header_block = cursor.document().findBlock(current_cell.start_position)
        sel_end_block = current_header_block.previous()
        if sel_end_block.isValid():
            sel_end = sel_end_block.position() + sel_end_block.length() - 1
        else:
            sel_end = target_pos

        return self._set_motion_info(
            target_pos,
            sel_start=target_pos,
            sel_end=sel_end,
            motion_type=MotionType.LineWise,
        )

    def next_cell(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return motion info for the next code cell."""
        cells = self.vim_status.get_cells()
        if len(cells) <= 1:
            return self._set_motion_info(None)

        cursor = self.get_cursor()
        cursor_pos = cursor.position()
        current_idx = self._locate_cell_index(cells, cursor_pos)
        if current_idx is None:
            return self._set_motion_info(None)

        target_idx = current_idx + num
        if target_idx >= len(cells):
            return self._set_motion_info(None)

        target_cell = cells[target_idx]
        target_pos = target_cell.start_position

        target_block = cursor.document().findBlock(target_pos)
        prev_block = target_block.previous()
        if prev_block.isValid():
            sel_end = prev_block.position() + prev_block.length() - 1
        else:
            sel_end = target_block.position()

        sel_start_block = cursor.block()
        sel_start = sel_start_block.position()

        if sel_end < sel_start:
            sel_end = sel_start

        return self._set_motion_info(
            target_pos,
            sel_start=sel_start,
            sel_end=sel_end,
            motion_type=MotionType.LineWise,
        )

    def i_cell(self, num: int = 1) -> MotionInfo:
        """Return motion info covering the current cell and subsequent cells."""
        cells = self.vim_status.get_cells()
        if not cells:
            return self._set_motion_info(None)

        cursor = self.get_cursor()
        cursor_pos = cursor.position()
        current_idx = self._locate_cell_index(cells, cursor_pos)
        if current_idx is None:
            return self._set_motion_info(None)

        count = max(1, num)
        end_idx = min(len(cells) - 1, current_idx + count - 1)

        document = cursor.document()
        start_cell = cells[current_idx]
        start_block = document.findBlockByNumber(start_cell.start_line)
        if not start_block.isValid():
            start_block = document.firstBlock()
        sel_start = start_block.position()

        end_cell = cells[end_idx]
        end_block = document.findBlockByNumber(end_cell.end_line)
        if not end_block.isValid():
            end_block = document.lastBlock()
        sel_end = end_block.position() + end_block.length() - 1
        if sel_end < sel_start:
            sel_end = sel_start

        return self._set_motion_info(
            sel_start,
            sel_start=sel_start,
            sel_end=sel_end,
            motion_type=MotionType.LineWise,
        )

    def _get_ch(self, pos):
        """Get character of position."""
        if pos >= self.vim_status.cursor.get_end_position():
            return ""
        cursor = self.get_cursor()
        cursor.setPosition(pos)
        cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
        return cursor.selectedText()

    def _get_leading_ch(self, pos):
        """Get character of position - 1."""
        cursor = self.get_cursor()
        cursor.setPosition(pos)
        cursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor)
        return cursor.selectedText()

    def zero(self, num=1, num_str=""):
        """Get the start position of the current line."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.StartOfLine)
        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def l(self, num=1, num_str=""):
        """Get the position on the right side of the cursor."""
        cursor = self.get_cursor()
        pos_cur = cursor.position()

        if cursor.atBlockStart() and cursor.atBlockEnd():
            return self._set_motion_info(pos_cur, motion_type=MotionType.CharWise)

        cursor.movePosition(QTextCursor.EndOfBlock)
        pos_end = cursor.position()

        pos_new = pos_cur + num
        if pos_new > pos_end:
            pos_new = pos_end

        return self._set_motion_info(pos_new, motion_type=MotionType.CharWise)

    def h(self, num=1, num_str=""):
        """Get the position on the left side of the cursor."""
        cursor = self.get_cursor()
        pos_cur = cursor.position()

        if cursor.atBlockStart() and cursor.atBlockEnd():
            return self._set_motion_info(pos_cur, motion_type=MotionType.CharWise)

        cursor.movePosition(QTextCursor.StartOfBlock)
        pos_start = cursor.position()

        pos_new = pos_cur - num
        if pos_new < pos_start:
            pos_new = pos_start

        return self._set_motion_info(pos_new, motion_type=MotionType.CharWise)

    def k(self, num=1, num_str=""):
        """Get the position above the cursor."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.Up, n=num)
        pos_new = cursor.position()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            pos_new -= 1

        return self._set_motion_info(pos_new)

    def j(self, num=1, num_str=""):
        """Get the position below the cursor."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.Down, n=num)
        pos_new = cursor.position()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            pos_new -= 1

        return self._set_motion_info(pos_new)

    def H(self, num=1, num_str=""):
        """Get the position of the top of page."""
        editor = self.get_editor()
        pos = editor.cursorForPosition(QPoint(0, 0)).position()
        return self._set_motion_info(pos)

    def M(self, num=1, num_str=""):
        """Get the position of the middle of page."""
        editor = self.get_editor()
        qpos_mid = int(editor.viewport().height() * 0.5)
        block = editor.cursorForPosition(QPoint(0, qpos_mid)).block()
        return self._set_motion_info(block.position())

    def L(self, num=1, num_str=""):
        """Get the position of the bottom of page."""
        editor = self.get_editor()
        qpos_mid = int(editor.viewport().height())
        block = editor.cursorForPosition(QPoint(0, qpos_mid)).block()
        return self._set_motion_info(block.position())

    def dollar(self, num=1, num_str=""):
        """Get the position of the end of the current line."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.EndOfLine, n=num)
        pos_new = cursor.position()

        return self._set_motion_info(pos_new, motion_type=MotionType.CharWise)

    def caret(self, num=1, num_str=""):
        """Get the position of the first non-blank character of the line."""
        cursor = self.get_cursor()
        block = cursor.block()
        text = block.text()
        pos = cursor.position()
        if text.strip():
            start_of_line = len(text) - len(text.lstrip())
            pos = block.position() + start_of_line

        return self._set_motion_info(pos, motion_type=MotionType.CharWise)

    def w(self, num=1, num_str=""):
        """Get the position of the next word."""
        cursor = self.get_cursor()
        for _ in range(num):
            cursor.movePosition(QTextCursor.NextWord)
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def w_for_d(self, num=1, num_str=""):
        """Get the position of the next word in d command."""
        cursor = self.get_cursor()
        for _ in range(num):
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.NextWord)
            else:
                cursor.movePosition(QTextCursor.NextWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def w_for_c(self, num=1, num_str=""):
        """Get the position of the next word for c command."""
        cursor = self.get_cursor()
        for _ in range(num - 1):
            cursor.movePosition(QTextCursor.NextWord)
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord)

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor.movePosition(QTextCursor.NextWord)
        else:
            cursor.movePosition(QTextCursor.EndOfWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def W(self, num=1, num_str=""):
        """Get the position of the next WORD."""
        cursor = self.get_cursor()
        for _ in range(num):
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord)
            else:
                cursor.movePosition(QTextCursor.NextWord)
                while self._get_leading_ch(cursor.position()) not in WHITE_SPACE:
                    if cursor.atBlockEnd():
                        cursor.movePosition(QTextCursor.NextWord)
                        if self._get_ch(cursor.position()) in WHITE_SPACE:
                            cursor.movePosition(QTextCursor.NextWord)
                        break
                    cursor.movePosition(QTextCursor.NextWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def W_for_d(self, num=1, num_str=""):
        """Get the position of the next WORD."""
        cursor = self.get_cursor()
        for _ in range(num):
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord)

            cursor.movePosition(QTextCursor.NextWord)
            while self._get_leading_ch(cursor.position()) not in WHITE_SPACE:
                if cursor.atBlockEnd():
                    break
                cursor.movePosition(QTextCursor.NextWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def W_for_c(self, num=1, num_str=""):
        """Get the position of the next WORD for c command."""
        cursor = self.get_cursor()
        for _ in range(num - 1):
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.NextWord)
                if self._get_ch(cursor.position()) in WHITE_SPACE:
                    cursor.movePosition(QTextCursor.NextWord)
            else:
                cursor.movePosition(QTextCursor.NextWord)
                while self._get_leading_ch(cursor.position()) not in WHITE_SPACE:
                    if cursor.atBlockEnd():
                        cursor.movePosition(QTextCursor.NextWord)
                        if self._get_ch(cursor.position()) in WHITE_SPACE:
                            cursor.movePosition(QTextCursor.NextWord)
                        break
                    cursor.movePosition(QTextCursor.NextWord)

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor.movePosition(QTextCursor.NextWord)
        else:
            cursor.movePosition(QTextCursor.EndOfWord)
            while (
                self._get_ch(cursor.position()) not in WHITE_SPACE
                and not cursor.atBlockEnd()
            ):
                cursor.movePosition(QTextCursor.Right)
                cursor.movePosition(QTextCursor.EndOfWord)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def b(self, num=1, num_str=""):
        """Get the position of the previous word."""
        cursor = self.get_cursor()

        def move2previousword(_cursor):
            _cursor.movePosition(QTextCursor.PreviousWord)
            while (
                cursor.atBlockStart() and self._get_ch(cursor.position()) in WHITE_SPACE
            ):
                if _cursor.position() == 0:
                    break
                if _cursor.atBlockStart() and cursor.atBlockEnd():
                    break
                _cursor.movePosition(QTextCursor.PreviousWord)
            return _cursor

        for _ in range(num):
            cursor = move2previousword(cursor)
            while cursor.atBlockEnd() and not cursor.atBlockStart():
                cursor = move2previousword(cursor)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def B(self, num=1, num_str=""):
        """Get the position of the previous WORD."""
        cursor = self.get_cursor()

        def move2previousWORD(_cursor):
            _cursor.movePosition(QTextCursor.PreviousWord)
            while self._get_leading_ch(cursor.position()) not in WHITE_SPACE:
                if _cursor.position() == 0:
                    break
                if _cursor.atBlockStart() and cursor.atBlockEnd():
                    break
                _cursor.movePosition(QTextCursor.PreviousWord)
            return _cursor

        for _ in range(num):
            cursor = move2previousWORD(cursor)
            while cursor.atBlockEnd() and not cursor.atBlockStart():
                cursor = move2previousWORD(cursor)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def e(self, num=1, num_str=""):
        """Get the position of the end of word.

        Does not stop in an empty line.

        """
        cursor = self.get_cursor()
        pos_old = cursor.position()

        if not cursor.atBlockEnd():
            cursor.movePosition(QTextCursor.Right)

        def _helper_e(cursor):
            while (
                cursor.atBlockEnd()
                or not cursor.block().text().strip()
                or self._get_ch(cursor.position()) in WHITE_SPACE
            ):
                cursor.movePosition(QTextCursor.NextWord)
                if cursor.atEnd():
                    break

            cursor.movePosition(QTextCursor.EndOfWord)
            return cursor

        for _ in range(num):
            cursor = _helper_e(cursor)

        if pos_old < cursor.position():
            cursor.movePosition(QTextCursor.Left)

        return self._set_motion_info(
            cursor.position(), motion_type=MotionType.CharWiseIncludingEnd
        )

    def G(self, num=1, num_str=""):
        """Get the position of the Line.

        if num_str is False, this function return the last line position.

        """
        editor = self.get_editor()
        n_block = editor.blockCount()

        if num > n_block:
            cursor = self.get_cursor()
            return self._set_motion_info(cursor.position())

        if num_str:
            block = editor.document().findBlockByNumber(num - 1)
        else:
            block = editor.document().findBlockByNumber(n_block - 1)

        text = block.text()
        start_of_line = len(text) - len(text.lstrip())
        if not text.strip() and text:
            start_of_line -= 1
        pos = block.position() + start_of_line

        return self._set_motion_info(pos)

    def percent(self, num=1, num_str=""):
        """Get the position of matching bracket."""
        txt = self.get_editor().toPlainText()
        cursor_pos = self.get_cursor().position()
        pos_start = None
        pos_end = None

        for match in re.finditer(r"[\(\[\{\}\]\)]", txt[cursor_pos::]):
            pos_start = match.start()
            ch_start = match.group()
            if ch_start in ")]}":
                pos_start += cursor_pos
                sub_txt = reversed(txt[:pos_start])
            else:
                pos_start += cursor_pos + 1
                sub_txt = txt[pos_start::]
            break

        if pos_start is None:
            return self._set_motion_info(None)

        stack = [ch_start]

        for idx, ch in enumerate(sub_txt):
            if ch == BRACKET_PAIR.get(stack[-1], None):
                stack.pop()
            elif ch in "()[]{}":
                stack.append(ch)
            if not stack and (ch_start in "([{"):
                pos_end = pos_start + idx
                break
            elif not stack and (ch_start in "}])"):
                pos_end = pos_start - idx - 1
                break

        return self._set_motion_info(
            pos_end, motion_type=MotionType.CharWiseIncludingEnd
        )

    def find_ch(self, ch, num=1, by_repeat_cmd=False):
        """Get the position of the next occurrence of a character."""
        if by_repeat_cmd is False:
            self.vim_status.find_info.set("f", ch)
        cursor = self.get_cursor()
        block = cursor.block()
        search_pos = cursor.positionInBlock() + 1
        txt = block.text()
        ch_pos = -1
        for _ in range(num):
            ch_pos = txt.find(ch, search_pos)
            if ch_pos < 0:
                break
            search_pos = ch_pos + 1

        if ch_pos < 0:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        else:
            return self._set_motion_info(
                ch_pos + block.position(), motion_type=MotionType.CharWiseIncludingEnd
            )

    def rfind_ch(self, ch, num=1, by_repeat_cmd=False):
        """Get the position of the previous occurrence of a character."""
        if by_repeat_cmd is False:
            self.vim_status.find_info.set("F", ch)
        cursor = self.get_cursor()
        block = cursor.block()
        search_pos = cursor.positionInBlock()
        txt = block.text()
        ch_pos = -1
        for _ in range(num):
            ch_pos = txt[:search_pos].rfind(ch)
            if ch_pos < 0:
                break
            search_pos = ch_pos

        if ch_pos < 0:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        else:
            return self._set_motion_info(
                ch_pos + block.position(), motion_type=MotionType.CharWise
            )

    def t(self, ch, num=1, by_repeat_cmd=False):
        """Get the position - 1 of the next occurrence of a character."""
        if by_repeat_cmd is False:
            self.vim_status.find_info.set("t", ch)
        cursor = self.get_cursor()
        block = cursor.block()
        search_pos = cursor.positionInBlock() + 1
        txt = block.text()
        ch_pos = -1
        for idx in range(num):
            if num == 1 and by_repeat_cmd is True:
                search_pos += 1
            ch_pos = txt.find(ch, search_pos)
            if ch_pos < 0:
                break
            search_pos = ch_pos + 1

        ch_pos -= 1
        if ch_pos < 0:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        else:
            return self._set_motion_info(
                ch_pos + block.position(), motion_type=MotionType.CharWiseIncludingEnd
            )

    def T(self, ch, num=1, by_repeat_cmd=False):
        """Get the position + 1 of the previous occurrence of a character."""
        if by_repeat_cmd is False:
            self.vim_status.find_info.set("T", ch)
        cursor = self.get_cursor()
        block = cursor.block()
        search_pos = cursor.positionInBlock()
        txt = block.text()
        ch_pos = -1
        for idx in range(num):
            if num == 1 and by_repeat_cmd is True:
                search_pos -= 1
                if search_pos == -1:
                    break
            ch_pos = txt[:search_pos].rfind(ch)
            if ch_pos < 0:
                break
            search_pos = ch_pos

        if ch_pos < 0:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)

        ch_pos += 1
        if ch_pos >= block.length():
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        return self._set_motion_info(
            ch_pos + block.position(), motion_type=MotionType.CharWise
        )

    def get_cursor_pos_of_viewport(self) -> tuple[int, int]:
        """Return start and end positions of the visible viewport."""
        return self.search_helper.get_viewport_positions()

    def search_forward_in_view(self, txt: str) -> list[int]:
        """Return positions of ``txt`` within the visible viewport."""
        return self.search_helper.search_forward_in_view(txt)

    def sneak(self, ch2, num=1, by_repeat_cmd=False):
        """Delegate to :class:`SearchHelper`."""
        return self.search_helper.sneak(ch2, num, by_repeat_cmd)

    def display_another_group_after_sneak(self):
        """Show annotations for additional sneak targets."""
        self.search_helper.display_another_group_after_sneak()

    def search_backward_in_view(self, txt: str) -> list[int]:
        """Return positions of ``txt`` when searching backward in the viewport."""
        return self.search_helper.search_backward_in_view(txt)

    def rsneak(self, ch2, num=1, by_repeat_cmd=False):
        """Delegate to :class:`SearchHelper`."""
        return self.search_helper.rsneak(ch2, num, by_repeat_cmd)

    def display_another_group_after_rsneak(self):
        """Show annotations for additional reverse sneak targets."""
        self.search_helper.display_another_group_after_rsneak()

    def semicolon(self, num=1, num_str=""):
        """Repeat latest f, t, F or T."""
        name = self.vim_status.find_info.cmd_name
        ch = self.vim_status.find_info.ch
        method = self.find_cmd_map.get(name, None)
        if method is None:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        motion_info = method(ch, num, True)
        return motion_info

    def comma(self, num=1, num_str=""):
        """Repeat latest f, t, F or T in opposite direction."""
        name = self.vim_status.find_info.cmd_name
        ch = self.vim_status.find_info.ch
        method = self.find_cmd_map.get(name.swapcase(), None)
        if method is None:
            return self._set_motion_info(None, motion_type=MotionType.CharWise)
        motion_info = method(ch, num, True)
        return motion_info

    def apostrophe(self, mark: str):
        """Return linewise position of mark."""
        info = self.vim_status.get_bookmark(mark)
        if not info:
            return self._set_motion_info(None)
        current = self.vim_status.get_editorstack().get_current_filename()
        if info.get("file") != current:
            return self._set_motion_info(None)
        editor = self.get_editor()
        block = editor.document().findBlockByNumber(info["line"])
        if not block.isValid():
            if mark.isupper():
                self.vim_status.bookmarks_global.pop(mark, None)
                self.vim_status._save_persistent_bookmarks()
            else:
                self.vim_status.bookmarks[current].pop(mark, None)
            return self._set_motion_info(None)
        return self._set_motion_info(block.position(), motion_type=MotionType.LineWise)

    def backtick(self, mark: str):
        """Return charwise position of mark."""
        info = self.vim_status.get_bookmark(mark)
        if not info:
            return self._set_motion_info(None)
        current = self.vim_status.get_editorstack().get_current_filename()
        if info.get("file") != current:
            return self._set_motion_info(None)
        editor = self.get_editor()
        block = editor.document().findBlockByNumber(info["line"])
        if not block.isValid():
            if mark.isupper():
                self.vim_status.bookmarks_global.pop(mark, None)
                self.vim_status._save_persistent_bookmarks()
            else:
                self.vim_status.bookmarks[current].pop(mark, None)
            return self._set_motion_info(None)
        pos = block.position() + min(info["col"], block.length() - 1)
        return self._set_motion_info(pos, motion_type=MotionType.CharWiseIncludingEnd)

    def i_quote(self, quote):
        """Find quoted block position excluding quotes."""
        sel_end = None

        cursor = self.get_cursor()
        block = cursor.block()
        text = block.text()

        cursor_pos = cursor.positionInBlock()
        if self._get_ch(cursor.position()) == quote:
            # When the cursor starts on a quote, Vim will figure out which
            # quote pairs form a string by searching from the start of the line
            quote_pos = [m.start() for m in re.finditer(quote, text[:cursor_pos])]
            if len(quote_pos) % 2 == 0:
                sel_start = cursor_pos
                sel_end = text.find(quote, cursor_pos + 1)
            else:
                sel_start = quote_pos[-1]
                sel_end = cursor_pos
        else:
            sel_start = text[:cursor_pos].rfind(quote)

            if sel_start < 0:
                sel_start = text.find(quote, cursor_pos)
                if sel_start < 0:
                    pass
                else:
                    sel_end = text.find(quote, sel_start + 1)
            else:
                sel_end = text.find(quote, cursor_pos)

        if sel_start > -1 and sel_end > -1:
            sel_start = sel_start + block.position() + 1
            sel_end = sel_end + block.position()
        else:
            sel_start = None
            sel_end = None

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def _include_trailing_white(self, pos_end):
        """Add the position of trailing white space."""
        cursor = self.get_cursor()

        cursor.setPosition(pos_end)

        # including trailing text
        if (
            pos_end < self.vim_status.cursor.get_end_position()
            and not cursor.atBlockEnd()
        ):
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
            ch = cursor.selectedText()

            cursor.setPosition(pos_end)
            if ch in WHITE_SPACE:
                cursor.movePosition(QTextCursor.NextWord)
                pos_end = cursor.position()

        return pos_end

    def _include_leading_white(self, pos_start):
        """Add the position of leading white."""
        cursor = self.get_cursor()

        cursor.setPosition(pos_start)
        block = cursor.block()
        text = block.text()
        text = text[: pos_start - block.position()]

        pos_start -= len(text) - len(text.rstrip())

        return pos_start

    def a_quote(self, quote):
        """Find quoted block position including trailing white space."""
        sel_end = None

        cursor = self.get_cursor()
        block = cursor.block()
        text = block.text()

        cursor_pos = cursor.positionInBlock()
        if self._get_ch(cursor.position()) == quote:
            # When the cursor starts on a quote, Vim will figure out which
            # quote pairs form a string by searching from the start of the line
            quote_pos = [m.start() for m in re.finditer(quote, text[:cursor_pos])]
            if len(quote_pos) % 2 == 0:
                sel_start = cursor_pos
                sel_end = text.find(quote, cursor_pos + 1)
            else:
                sel_start = quote_pos[-1]
                sel_end = cursor_pos
        else:
            sel_start = text[:cursor_pos].rfind(quote)

            if sel_start < 0:
                sel_start = text.find(quote, cursor_pos)
                if sel_start < 0:
                    pass
                else:
                    sel_end = text.find(quote, sel_start + 1)
            else:
                sel_end = text.find(quote, cursor_pos)

        if sel_start > -1 and sel_end > -1:
            sel_start = sel_start + block.position()
            sel_end = sel_end + block.position() + 1
            sel_end_old = sel_end
            sel_end = self._include_trailing_white(sel_end)
            if sel_end == sel_end_old:
                sel_start = self._include_leading_white(sel_start)
        else:
            sel_start = None
            sel_end = None

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def i_w(self, num):
        """Find word block position."""
        cursor = self.get_cursor()

        sel_start_init = cursor.position()

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor_pos_in_block = cursor.positionInBlock()
            txt = cursor.block().text()
            txt_leading = txt[:cursor_pos_in_block]
            txt_trailing = txt[cursor_pos_in_block:]

            n_leading_space = len(txt_leading) - len(txt_leading.rstrip())
            n_trailing_space = len(txt_trailing) - len(txt_trailing.lstrip())
            sel_start = cursor_pos_in_block - n_leading_space
            sel_end = cursor_pos_in_block + n_trailing_space
            cursor.setPosition(sel_end + cursor.block().position())
            num -= 1
        else:
            cursor.movePosition(QTextCursor.EndOfWord)
            cursor.movePosition(QTextCursor.StartOfWord)

            # For "abc.dd" at 0 position of cursor
            if sel_start_init < cursor.position():
                cursor.movePosition(QTextCursor.Left)
                cursor.movePosition(QTextCursor.StartOfWord)

            sel_start = cursor.position()

        for idx in range(num):
            if idx % 2 == 0:
                cursor.movePosition(QTextCursor.EndOfWord)
            else:
                # go to the end of blank
                if cursor.atBlockEnd():
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.NextWord)

        sel_end = cursor.position()

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def i_W(self, num):
        """Find WORD block position."""
        cursor = self.get_cursor()

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor_pos_in_block = cursor.positionInBlock()
            txt = cursor.block().text()
            txt_leading = txt[:cursor_pos_in_block]
            txt_trailing = txt[cursor_pos_in_block:]

            n_leading_space = len(txt_leading) - len(txt_leading.rstrip())
            n_trailing_space = len(txt_trailing) - len(txt_trailing.lstrip())
            sel_start = cursor_pos_in_block - n_leading_space
            sel_end = cursor_pos_in_block + n_trailing_space
            cursor.setPosition(sel_end + cursor.block().position())
            num -= 1
        else:
            cursor.movePosition(QTextCursor.EndOfWord)
            cursor.movePosition(QTextCursor.StartOfWord)

            while (
                not cursor.atBlockStart()
                and self._get_leading_ch(cursor.position()) not in WHITE_SPACE
            ):
                cursor.movePosition(QTextCursor.Left)
                cursor.movePosition(QTextCursor.StartOfWord)

            sel_start = cursor.position()

        for idx in range(num):
            if idx % 2 == 0:
                if cursor.atBlockEnd():
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.EndOfWord)
                    while (
                        not cursor.atBlockEnd()
                        and self._get_ch(cursor.position()) not in WHITE_SPACE
                    ):
                        cursor.movePosition(QTextCursor.Right)
                        if self._get_ch(cursor.position()) not in WHITE_SPACE:
                            cursor.movePosition(QTextCursor.EndOfWord)
            else:
                # go to the end of blank
                if cursor.atBlockEnd():
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.NextWord)

        sel_end = cursor.position()

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def a_w(self, num):
        """Find word block position including white space."""
        cursor = self.get_cursor()

        sel_start_init = cursor.position()

        is_trailing_space = True

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor_pos_in_block = cursor.positionInBlock()
            txt = cursor.block().text()
            txt_leading = txt[:cursor_pos_in_block]

            n_leading_space = len(txt_leading) - len(txt_leading.rstrip())
            sel_start = cursor_pos_in_block - n_leading_space
            cursor.movePosition(QTextCursor.NextWord)
            cursor.movePosition(QTextCursor.EndOfWord)
            is_trailing_space = False
            num -= 1
        else:
            cursor.movePosition(QTextCursor.EndOfWord)
            cursor.movePosition(QTextCursor.StartOfWord)

            # For "abc.dd" at 0 position of cursor
            if sel_start_init < cursor.position():
                cursor.movePosition(QTextCursor.Left)
                cursor.movePosition(QTextCursor.StartOfWord)

            sel_start = cursor.position()

        if is_trailing_space:
            for idx in range(num):
                if cursor.atBlockEnd():
                    # TODO : Fix the case that the line has leading space.
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.NextWord)
        else:
            for idx in range(num):
                block = cursor.block()
                txt = block.text()
                if len(txt.rstrip()) == cursor.position() - block.position():
                    cursor.movePosition(QTextCursor.NextBlock)
                    cursor.movePosition(QTextCursor.StartOfLine)
                    if self._get_ch(cursor.position()) in WHITE_SPACE:
                        cursor.movePosition(QTextCursor.NextWord)
                    cursor.movePosition(QTextCursor.EndOfWord)
                else:
                    cursor.movePosition(QTextCursor.NextWord)
                    cursor.movePosition(QTextCursor.EndOfWord)

        sel_end = cursor.position()

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def get_pos_bracket(
        self, num: int, bracket: str, cursor_pos: int
    ) -> tuple[int | None, int | None]:
        """Get the position of the bracket block."""
        # Todo: apply num
        pair_bracket = {
            "(": ["(", ")", r"[()]"],
            ")": ["(", ")", r"[()]"],
            "[": ["[", "]", r"[\[\]]"],
            "]": ["[", "]", r"[\[\]]"],
            "{": ["{", "}", r"[{}]"],
            "}": ["{", "}", r"[{}]"],
        }
        b_open, b_close, pattern_str = pair_bracket[bracket]
        txt = self.get_editor().toPlainText()

        sel_start = None
        sel_end = None

        pattern = re.compile(pattern_str)

        # Find position of close bracket
        if self._get_ch(cursor_pos) == b_close:
            sel_end = cursor_pos + 1
        else:
            n_open = 0
            for match in pattern.finditer(txt[cursor_pos + 1 : :]):
                ch = match.group()
                if ch == b_open:
                    n_open += 1
                else:
                    if n_open > 0:
                        n_open -= 1
                    else:
                        sel_end = match.start() + cursor_pos + 1 + 1
                        break

        # Find position of open bracket
        if sel_end is None:
            pass
        elif self._get_ch(cursor_pos) == b_open:
            sel_start = cursor_pos
        else:
            n_close = 0
            for idx, ch in enumerate(reversed(txt[:cursor_pos])):
                if ch == b_close:
                    n_close += 1
                elif ch == b_open:
                    if n_close > 0:
                        n_close -= 1
                    else:
                        sel_start = cursor_pos - idx - 1
                        break

        if sel_start is None or sel_end is None:
            sel_start = None
            sel_end = None

        return sel_start, sel_end

    def a_bracket(self, num: int, bracket: str) -> MotionInfo:
        """Get the position of the bracket block."""
        cursor = self.get_cursor()
        cursor_pos = cursor.position()
        sel_start, sel_end = self.get_pos_bracket(num, bracket, cursor_pos)

        if sel_start is None:
            # There are an open bracket after the cursor on the same line.
            open_bracket = {"(": "(", ")": "(", "[": "[", "]": "[", "{": "{", "}": "{"}[
                bracket
            ]
            block = cursor.block()
            txt = block.text()
            block_pos = block.position()

            txt_right = txt[cursor_pos - block_pos :]
            open_bracket_pos = txt_right.find(open_bracket) + cursor_pos
            if open_bracket_pos > cursor_pos:
                sel_start, sel_end = self.get_pos_bracket(
                    num, bracket, open_bracket_pos
                )

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    def i_bracket(self, num: int, bracket: str) -> MotionInfo:
        """Get the position of the bracket inner block."""
        motion_info = self.a_bracket(num, bracket)
        if motion_info.sel_start is not None and motion_info.sel_end is not None:
            motion_info.sel_start += 1
            motion_info.sel_end -= 1
            if motion_info.sel_start == motion_info.sel_end:
                motion_info.sel_start = None
                motion_info.sel_end = None

        return motion_info

    def search(self, txt: str):
        """Delegate search to :class:`SearchHelper`."""
        self.search_helper.search(txt)

    def n(self, num=1, num_str=""):
        """Move to the next search match."""
        return self.search_helper.n(num, num_str)

    def N(self, num=1, num_str=""):
        """Move to the previous search match."""
        return self.search_helper.N(num, num_str)

    def _get_word_under_cursor(self) -> str:
        """Return word under cursor."""
        editor = self.get_editor()
        return editor.get_current_word()

    def asterisk(self, num=1, num_str=""):
        """Search forward for the word under the cursor."""
        return self.search_helper.asterisk(num, num_str)

    def sharp(self, num=1, num_str=""):
        """Search backward for the word under the cursor."""
        return self.search_helper.sharp(num, num_str)

    def space(self, num=1, num_str=""):
        """Get the position on the right side of the cursor."""
        cursor = self.get_cursor()

        for _ in range(num):
            cursor.movePosition(QTextCursor.Right)
            if cursor.atBlockEnd():
                cursor.movePosition(QTextCursor.Right)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def backspace(self, num=1, num_str=""):
        """Get the position on the right side of the cursor."""
        cursor = self.get_cursor()

        cursor.movePosition(QTextCursor.Left, n=num)

        return self._set_motion_info(cursor.position(), motion_type=MotionType.CharWise)

    def enter(self, num=1, num_str=""):
        """Get the position below the cursor."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.Down, n=num)
        block = cursor.block()
        text = block.text()
        pos = cursor.position()

        if text.strip():
            start_of_line = len(text) - len(text.lstrip())
            pos = block.position() + start_of_line

        return self._set_motion_info(pos)

    def s(self, num):
        """Line block position."""
        cursor = self.get_cursor()

        sel_start_init = cursor.position()

        if self._get_ch(cursor.position()) in WHITE_SPACE:
            cursor_pos_in_block = cursor.positionInBlock()
            txt = cursor.block().text()
            txt_leading = txt[:cursor_pos_in_block]
            txt_trailing = txt[cursor_pos_in_block:]

            n_leading_space = len(txt_leading) - len(txt_leading.rstrip())
            n_trailing_space = len(txt_trailing) - len(txt_trailing.lstrip())
            sel_start = cursor_pos_in_block - n_leading_space
            sel_end = cursor_pos_in_block + n_trailing_space
            cursor.setPosition(sel_end + cursor.block().position())
            num -= 1
        else:
            cursor.movePosition(QTextCursor.EndOfWord)
            cursor.movePosition(QTextCursor.StartOfWord)

            # For "abc.dd" at 0 position of cursor
            if sel_start_init < cursor.position():
                cursor.movePosition(QTextCursor.Left)
                cursor.movePosition(QTextCursor.StartOfWord)

            sel_start = cursor.position()

        for idx in range(num):
            if idx % 2 == 0:
                cursor.movePosition(QTextCursor.EndOfWord)
            else:
                # go to the end of blank
                if cursor.atBlockEnd():
                    cursor.movePosition(QTextCursor.NextWord, n=2)
                else:
                    cursor.movePosition(QTextCursor.NextWord)

        sel_end = cursor.position()

        return self._set_motion_info(
            None, sel_start, sel_end, motion_type=MotionType.BlockWise
        )

    # ------------------------------------------------------------------
    # Python navigation helpers
    # ------------------------------------------------------------------
    def _get_python_definition_positions(self) -> list[int]:
        """Return character positions of Python class and function definitions."""
        editor = self.get_editor()
        text = editor.toPlainText()

        try:
            tree = ast.parse(text)
        except SyntaxError:
            return []

        lines: set[int] = set()
        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
                lines.add(node.lineno)

        positions: list[int] = []
        for line in sorted(lines):
            block = editor.document().findBlockByNumber(line - 1)
            if not block.isValid():
                continue
            txt = block.text()
            start_of_line = len(txt) - len(txt.lstrip())
            positions.append(block.position() + start_of_line)

        return positions

    def prev_pydef(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return position of previous Python definition."""
        positions = self._get_python_definition_positions()
        cursor = self.get_cursor()
        cur_line_start = cursor.block().position()
        prev_positions = [p for p in positions if p < cur_line_start]
        if not prev_positions:
            return self._set_motion_info(None)
        idx = max(0, len(prev_positions) - num)
        return self._set_motion_info(prev_positions[idx], motion_type=MotionType.LineWise)

    def next_pydef(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return position of next Python definition."""
        positions = self._get_python_definition_positions()
        cursor = self.get_cursor()
        cur_line_start = cursor.block().position()
        next_positions = [p for p in positions if p > cur_line_start]
        if not next_positions:
            return self._set_motion_info(None)
        idx = min(len(next_positions) - 1, num - 1)
        return self._set_motion_info(next_positions[idx], motion_type=MotionType.LineWise)

    # ------------------------------------------------------------------
    # Python block navigation helpers
    # ------------------------------------------------------------------
    def _get_python_block_positions(self) -> list[int]:
        """Return start positions of Python code blocks."""
        editor = self.get_editor()
        doc = editor.document()
        positions: list[int] = []

        block_keywords = (
            "def ",
            "async def ",
            "class ",
            "if ",
            "for ",
            "async for ",
            "while ",
            "try:",
            "with ",
            "async with ",
        )

        for line_no, text in enumerate(editor.toPlainText().splitlines()):
            stripped = text.lstrip()
            if any(stripped.startswith(kw) for kw in block_keywords):
                block = doc.findBlockByNumber(line_no)
                if not block.isValid():
                    continue
                start_of_line = len(text) - len(stripped)
                positions.append(block.position() + start_of_line)

        return positions

    def prev_pyblock(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return position of previous Python block."""
        positions = self._get_python_block_positions()
        cursor = self.get_cursor()
        cur_line_start = cursor.block().position()
        prev_positions = [p for p in positions if p < cur_line_start]
        if not prev_positions:
            return self._set_motion_info(None)
        idx = max(0, len(prev_positions) - num)
        return self._set_motion_info(prev_positions[idx], motion_type=MotionType.LineWise)

    def next_pyblock(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Return position of next Python block."""
        positions = self._get_python_block_positions()
        cursor = self.get_cursor()
        cur_line_start = cursor.block().position()
        next_positions = [p for p in positions if p > cur_line_start]
        if not next_positions:
            return self._set_motion_info(None)
        idx = min(len(next_positions) - 1, num - 1)
        return self._set_motion_info(next_positions[idx], motion_type=MotionType.LineWise)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""OkVim Widget."""

# Standard Libraries
import os.path as osp
import sys
import threading
from functools import wraps

# Third Party Libraries
from qtpy import PYSIDE2, PYSIDE6
from qtpy.QtCore import QObject, Qt, QThread, Signal, Slot
from qtpy.QtGui import QFocusEvent, QKeyEvent, QKeySequence, QTextCursor
from qtpy.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QWidget, QStyle
from spyder.api.config.decorators import on_conf_change
from spyder.api.plugins import Plugins
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.executor import (
    ExecutorLeaderKey,
    ExecutorNormalCmd,
    ExecutorVisualCmd,
    ExecutorVlineCmd,
)
from spyder_okvim.spyder.config import CONF_SECTION, KEYCODE2STR
from spyder_okvim.utils.file_search import FileSearchDialog
from spyder_okvim.utils.testing_env import running_in_pytest
from spyder_okvim.vim import InputCmdInfo, KeyInfo, VimState, VimStatus

running_coverage = "coverage" in sys.modules


if PYSIDE2 or PYSIDE6:
    try:
        # Third Party Libraries
        from shiboken6 import isValid as _sbk_is_valid  # PySide6
    except Exception:
        try:
            # Third Party Libraries
            from shiboken2 import isValid as _sbk_is_valid  # PySide2
        except Exception:
            _sbk_is_valid = None

    def _widget_is_deleted(obj):
        if _sbk_is_valid is None:
            return False
        return not _sbk_is_valid(obj)
else:  # PYQT5
    try:
        # Third Party Libraries
        from PyQt5 import sip
    except Exception:  # pragma: no cover - fallback for standalone sip module
        try:
            # Third Party Libraries
            import sip  # type: ignore
        except Exception:
            sip = None

    def _widget_is_deleted(obj):
        if sip is None:
            return False
        return sip.isdeleted(obj)


def enable_coverage_tracing(fn):
    """Apply coverage tracing to a thread entry point."""

    @wraps(fn)
    def wrapped(*args, **kwargs):
        if running_coverage:
            sys.settrace(threading._trace_hook)
        fn(*args, **kwargs)

    return wrapped


class VimPane(PluginMainWidget):
    """Invisible pane used to host the Vim command widget."""

    def __init__(self, name=None, plugin=None, parent=None):
        """Create the pane and its underlying Vim widget."""
        super().__init__(name, plugin, parent)
        layout = QHBoxLayout()
        layout.addWidget(QLabel("Please hide this pane."))
        self.setLayout(layout)

        self.main = parent
        self.vim_cmd = VimWidget(self.main.editor, self.main)

    def get_title(self):
        """Return the localized title for the pane."""
        return "Okvim"

    def get_focus_widget(self):
        """Return the widget that should receive focus."""
        return self.vim_cmd

    def setup(self):
        """Perform initial setup after the plugin is created."""
        pass

    def update_actions(self):
        """Update actions exposed by the plugin."""
        pass

    @on_conf_change
    def apply_plugin_settings(self, options):
        """Apply the config settings."""
        self.vim_cmd.vim_status.search.set_color()
        self.vim_cmd.vim_status.cursor.set_config_from_conf()
        self.vim_cmd.set_leader_key()


class VimShortcut(QObject):
    """Method class for vim shortcut."""

    signal_cmd = Signal(str)

    def __init__(self, main, vim_status: VimStatus):
        super().__init__()
        self.main = main
        self.vim_status = vim_status
        self.get_editor = self.vim_status.get_editor
        self.cmd_line = None

    def _scroll(self, half: bool, up: bool) -> None:
        """Scroll the editor window.

        Args:
            half: If ``True`` scroll half the visible lines.
            up: Scroll up when ``True`` otherwise scroll down.
        """
        lines = self.vim_status.get_number_of_visible_lines()
        if half:
            lines //= 2
        if lines < 1:
            lines = 1

        command = "k" if up else "j"
        self.signal_cmd.emit(f"{lines}{command}")
        self.signal_cmd.emit("^")

    def pg_half_up(self):
        """Scroll half page up."""
        self._scroll(True, True)

    def pg_up(self):
        """Scroll one page up."""
        self._scroll(False, True)

    def pg_half_down(self):
        """Scroll half page down."""
        self._scroll(True, False)

    def pg_down(self):
        """Scroll one page down."""
        self._scroll(False, False)

    def _extract_number(self) -> tuple[int | None, int, int]:
        """Return the number under the cursor.

        Returns:
            tuple[int | None, int, int]: Extracted number and its start and end
            positions.
        """
        cursor = self.get_editor().textCursor()

        val = None
        block_start = cursor.block().position()
        block_end = block_start + cursor.block().length()
        pos_start = cursor.position()
        pos_end = pos_start

        for index_end in range(pos_start + 1, block_end):
            cursor.setPosition(pos_start)
            cursor.setPosition(index_end, QTextCursor.KeepAnchor)
            try:
                text_selected = cursor.selectedText()
                if text_selected == "-":
                    continue
                val_ = int(text_selected)
                if text_selected.isdigit() or text_selected[0] == "-":
                    val = val_
                    pos_end = index_end
                else:
                    break
            except ValueError:
                break

        for index_start in range(pos_start, block_start - 1, -1):
            cursor.setPosition(index_start)
            cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
            try:
                val_ = int(cursor.selectedText())
                text_selected = cursor.selectedText()
                if text_selected.isdigit() or text_selected[0] == "-":
                    val = val_
                    pos_start = index_start
                else:
                    break
            except ValueError:
                break

        return val, pos_start, pos_end

    def _change_number(self, delta: int, key: int) -> None:
        """Change the number at the cursor.

        Args:
            delta: Increment or decrement value.
            key: Qt key code used for dot command updates.
        """
        val, start, end = self._extract_number()
        if val is None:
            return
        if self.vim_status.sub_mode:
            self.cmd_line.esc_pressed()
            return

        count_text = self.cmd_line.text()
        count = 1 if not count_text else int(count_text)

        cursor = self.get_editor().textCursor()
        cursor.setPosition(start)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        cursor.insertText(str(val + delta * count))
        self.signal_cmd.emit("h")

        cmd_info = InputCmdInfo(str(count), "")
        self.vim_status.input_cmd.set(cmd_info)
        key_info = KeyInfo(key, "", Qt.ControlModifier, 0)
        self.vim_status.update_dot_cmd(False, key_list_to_cmd_line=[key_info])

    def add_num(self) -> None:
        """Add to the number at the cursor."""
        self._change_number(1, Qt.Key_A)

    def subtract_num(self) -> None:
        """Subtract from the number at the cursor."""
        self._change_number(-1, Qt.Key_X)

    def redo(self) -> None:
        """Redo [count] changes which were undone."""
        if self.vim_status.sub_mode:
            self.cmd_line.esc_pressed()
            return

        if not self.vim_status.is_normal():
            return

        editor = self.vim_status.get_editor()
        n_block_old = editor.blockCount()

        txt = self.cmd_line.text()
        num = 1 if not txt else int(txt)

        editor = self.get_editor()
        for _ in range(num):
            editor.redo()

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            if n_block_new > n_block_old:
                self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")
            else:
                self.vim_status.set_message(f"{n_block_old - n_block_new} fewer lines")
        else:
            self.vim_status.set_message(f"{num} changes")

        self.cmd_line.esc_pressed()

    def open_file_search(self) -> None:
        """Open the file search dialog."""
        root_folder = self.main.projects.get_active_project_path()

        dlg = FileSearchDialog(root_folder, self.main)
        dlg.exec_()
        path = dlg.get_selected_path()

        if osp.isfile(path):
            self.vim_status.push_jump()
            application = self.main.get_plugin(Plugins.Application)
            application.open_file_in_plugin(path)
            self.vim_status.jump_list.push(path, 0)
            self.vim_status.set_focus_to_vim()

    def clear_tip_search(self) -> None:
        """Clear tooltip, search highlight."""
        self.get_editor().hide_tooltip()
        self.vim_status.cursor.set_extra_selections("vim_search", [])

    def jump_backward(self) -> None:
        """Jump to previous location in the jumplist."""
        if self.vim_status.sub_mode:
            self.cmd_line.esc_pressed()
            return
        self.vim_status.jump_backward()
        self.vim_status.set_focus_to_vim()

    def jump_forward(self) -> None:
        """Jump to next location in the jumplist."""
        if self.vim_status.sub_mode:
            self.cmd_line.esc_pressed()
            return
        self.vim_status.jump_forward()
        self.vim_status.set_focus_to_vim()


class VimStateLabel(QLabel):
    """Display state of vim."""

    def __init__(self, parent):
        super().__init__(parent)
        self.change_state(VimState.INSERT)

        self.setAlignment(Qt.AlignCenter)
        tw = self.fontMetrics().width(" NORMAL ")
        fw = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)

    @Slot(int)
    def change_state(self, state):
        """Display the state of vim."""
        try:
            if _widget_is_deleted(self):
                return
            self.setStyleSheet("QLabel { color: white }")
            if state == VimState.VISUAL:
                self.setText("VISUAL")
                self.setStyleSheet("QLabel { background-color: #ff8000 }")
            elif state == VimState.NORMAL:
                self.setText("NORMAL")
                self.setStyleSheet("QLabel { background-color: #29a329 }")
            elif state == VimState.VLINE:
                self.setText("V-LINE")
                self.setStyleSheet("QLabel { background-color: #ff8000 }")
            elif state == VimState.INSERT:
                self.setText("INSERT")
                self.setStyleSheet("QLabel { background-color: #3366ff }")
        except RuntimeError:
            # Ignore updates to deleted widgets
            pass


class VimLineEdit(QLineEdit):
    """Vim Command input."""

    def __init__(self, vim_widget, vim_status: VimStatus, vim_shortcut: VimShortcut):
        super().__init__(vim_widget)
        self.vim_widget = vim_widget
        self.vim_status = vim_status
        self.vim_shortcut = vim_shortcut

        # Set size
        tw = self.fontMetrics().width(" :%s/international/internationl/g ")
        fw = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)

        # Todo: Move setting shortcut to config file.
        vim_shortcut.signal_cmd.connect(self.setText)
        self.dispatcher = {
            Qt.Key_A: vim_shortcut.add_num,
            Qt.Key_X: vim_shortcut.subtract_num,
            Qt.Key_D: vim_shortcut.pg_half_down,
            Qt.Key_F: vim_shortcut.pg_down,
            Qt.Key_U: vim_shortcut.pg_half_up,
            Qt.Key_B: vim_shortcut.pg_up,
            Qt.Key_R: vim_shortcut.redo,
            Qt.Key_P: vim_shortcut.open_file_search,
            Qt.Key_C: vim_shortcut.clear_tip_search,
            Qt.Key_O: vim_shortcut.jump_backward,
            Qt.Key_I: vim_shortcut.jump_forward,
        }
        self.setAttribute(Qt.WA_InputMethodEnabled, False)

    def to_normal(self) -> None:
        """Convert the state of vim to normal mode."""
        self.clear()
        self.vim_widget.vim_status.to_normal()
        self.vim_widget.vim_status.cursor.draw_vim_cursor()

    def keyPressEvent(self, e: QKeyEvent) -> None:
        """Override Qt method."""
        self.vim_status.manager_macro.add_vim_keyevent(e)

        key = e.key()
        pressed_ctrl = e.modifiers() == Qt.ControlModifier
        if key == Qt.Key_Escape:
            self.esc_pressed()
        elif KEYCODE2STR.get(key, None):
            self.setText(self.text() + KEYCODE2STR[key])
        elif pressed_ctrl and key in self.dispatcher.keys():
            self.dispatcher[key]()
        else:
            super().keyPressEvent(e)

    def esc_pressed(self) -> None:
        """Clear state."""
        self.vim_status.input_cmd.clear()
        self.vim_status.remove_marker_of_easymotion()
        if self.vim_status.sub_mode:
            self.vim_status.sub_mode = None
            self.clear()
        else:
            self.to_normal()

    def focusInEvent(self, event: QFocusEvent) -> None:
        """Override Qt method."""
        self.vim_status.disconnect_from_editor()
        super().focusInEvent(event)
        if self.vim_status.cursor.get_editor():
            self.to_normal()
        self.vim_status.set_message("")

    def focusOutEvent(self, event: QFocusEvent) -> None:
        """Override Qt method."""
        super().focusOutEvent(event)
        self.clear()
        try:
            self.vim_widget.vim_status.to_insert()
        except RuntimeError:
            # The status may already be deleted when closing widgets
            pass


class VimMessageLabel(QLabel):
    """Display Vim status messages."""

    def __init__(self, txt, parent):
        super().__init__(txt, parent)
        self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        tw = self.fontMetrics().width(" recording @q... 0000 fewers lines ")
        fw = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)


class MacroPlaybackWorker(QThread):
    """Replay a recorded macro on the main thread."""

    sig_focus_vim = Signal()
    sig_send_key_info = Signal(object)

    def __init__(self, parent: QObject | None) -> None:
        super().__init__(parent)
        self.key_info_list = []
        self.num_iteration = 0

    def set_key_infos(self, key_infos: list[KeyInfo], num: int) -> None:
        """Set key infos of registers."""
        self.key_info_list = key_infos.copy()
        self.num_iteration = num

    @enable_coverage_tracing
    def run(self) -> None:
        """Send key info to main thread."""
        is_focus_vim = True
        for _ in range(self.num_iteration):
            for key_info in self.key_info_list:
                if key_info.identifier == 1:
                    is_focus_vim = False
                elif key_info.identifier == 0 and is_focus_vim is False:
                    is_focus_vim = True
                    self.sig_focus_vim.emit()
                self.sig_send_key_info.emit(key_info)

        if is_focus_vim is False:
            self.sig_focus_vim.emit()


class VimWidget(QWidget):
    """Vim widget."""

    def __init__(self, editor_widget, main):
        super().__init__(main)
        self.editor_widget = editor_widget
        self.main = main
        self.status_label = VimStateLabel(main)
        self.msg_label = VimMessageLabel("", main)

        self.vim_status = VimStatus(editor_widget, main, self.msg_label)
        # # Avoid Qt crashes if the label is deleted by ignoring the signal
        # self.vim_status.change_label.connect(lambda state: None)
        self.vim_status.change_label.connect(self.status_label.change_state)

        self.vim_shortcut = VimShortcut(self.main, self.vim_status)

        self.commandline = VimLineEdit(self, self.vim_status, self.vim_shortcut)
        self.commandline.textChanged.connect(self.on_text_changed)

        self.vim_status.cmd_line = self.commandline
        self.vim_shortcut.cmd_line = self.commandline

        self.executor_normal_cmd = ExecutorNormalCmd(self.vim_status)
        self.executor_visual_cmd = ExecutorVisualCmd(self.vim_status)
        self.executor_vline_cmd = ExecutorVlineCmd(self.vim_status)
        self.executors = {
            VimState.NORMAL: self.executor_normal_cmd,
            VimState.VISUAL: self.executor_visual_cmd,
            VimState.VLINE: self.executor_vline_cmd,
        }

        # leader key
        self.executor_leader_key = ExecutorLeaderKey(self.vim_status)
        self.leader_key = " "
        self.set_leader_key()

        # macro
        self.worker_macro = MacroPlaybackWorker(main)
        self.worker_macro.sig_send_key_info.connect(self.send_key_event)
        self.worker_macro.sig_focus_vim.connect(self.commandline.setFocus)

    @Slot(object)
    def send_key_event(self, key_info: KeyInfo) -> None:
        event = key_info.to_event()
        if key_info.identifier == 0:
            self.commandline.keyPressEvent(event)
        else:
            editor = self.vim_status.get_editor()
            editor.keyPressEvent(event)

    def set_leader_key(self) -> None:
        """Set leader key from CONF."""
        leader_key = CONF.get(CONF_SECTION, "leader_key")
        leader_key2 = KEYCODE2STR.get(QKeySequence.fromString(leader_key)[0], None)
        if leader_key2:
            leader_key = leader_key2
        self.leader_key = leader_key
        self.executor_leader_key.set_easymotion_key(self.leader_key)

    def on_text_changed(self, txt: str) -> None:
        """Send input command to executor."""
        if not txt:
            return

        executor = self.executors[self.vim_status.vim_state]
        if self.vim_status.sub_mode:
            executor = self.vim_status.sub_mode

        # workaround for easymotion (press leader, leader)
        if txt == self.leader_key and executor != self.executor_leader_key:
            sub_mode = self.vim_status.sub_mode
            if sub_mode and sub_mode.allow_leaderkey is False:
                pass
            else:
                self.executor_leader_key.prev_executor = executor
                self.vim_status.sub_mode = self.executor_leader_key
                self.commandline.clear()
                return

        if executor(txt):
            self.commandline.clear()

        if self.vim_status.manager_macro.reg_name_for_execute:
            mm = self.vim_status.manager_macro
            ch = mm.reg_name_for_execute
            self.worker_macro.set_key_infos(mm.registers[ch], mm.num_execute)
            self.worker_macro.start()
            mm.set_info_for_execute("", 0)

    def cleanup(self) -> None:
        """Clean up resources used by the widget."""
        try:
            self.vim_status.change_label.disconnect(self.status_label.change_state)
        except Exception:
            pass
        if self.worker_macro.isRunning():
            self.worker_macro.quit()
            self.worker_macro.wait()
        if running_in_pytest():
            self.worker_macro.deleteLater()
            self.commandline.deleteLater()
            self.status_label.deleteLater()
            self.msg_label.deleteLater()



"""Tests for the executor_surround."""

# Third Party Libraries
import pytest

# Project Libraries
from spyder_okvim.vim import VimState


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["v", "S", "'"], "''", 0),
        ("", ["v", "S", "i"], "", 0),
        ("abde", ["l", "v", "S", "'"], "a'b'de", 1),
        ("abde", ["l", "v", "l" "S", '"'], 'a"bd"e', 1),
        ("abde", ["l", "v", "l" "S", "["], "a[ bd ]e", 1),
        ("abde", ["l", "v", "l" "S", "{"], "a{ bd }e", 1),
        ("abde", ["l", "v", "l" "S", "("], "a( bd )e", 1),
        ("abde", ["l", "v", "l" "S", "]"], "a[bd]e", 1),
        ("abde", ["l", "v", "l" "S", "}"], "a{bd}e", 1),
        ("abde", ["l", "v", "l" "S", ")"], "a(bd)e", 1),
        ("abde", ["l", "v", "l" "S", "B"], "a{bd}e", 1),
        ("abde", ["l", "v", "l" "S", "b"], "a(bd)e", 1),
        ("abde abcd", ["l", "v", "l" "S", ")", "6l", "."], "a(bd)e (ab)cd", 7),
        ("abde\ndd", ["l", "v", "j" "S", ")"], "a(bde\ndd)", 1),
    ],
)
def test_add_surround_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test to add surround command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")

    assert sel == []
    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected",
    [
        ("a", ["y", "s", "l", "("], 0, "( a )"),
        ("abcd", ["$", "y", "s", "2h", ")"], 1, "a(bc)d"),
        ("abcd", ["$", "y", "s", "2h", "b"], 1, "a(bc)d"),
        ("abcd", ["y", "s", "$", '"'], 0, '"abcd"'),
        ("abcd", ["$", "ys", "^", "'"], 0, "'abc'd"),
        ("a b", ["ys", "w", "}"], 0, "{a} b"),
        ("a b", ["ys", "w", "B"], 0, "{a} b"),
        ("a b", ["ys", "w", "}", "4l", "."], 4, "{a} {b}"),
        ("a b", ["l", "ys", "w", "("], 1, "a(   )b"),
        ("a b c ", ["ys", "2w", "]"], 0, "[a b] c "),
        ("a.dk b", ["ys", "W", "["], 0, "[ a.dk ] b"),
        ("a.dk b dd", ["ys", "2W", '"'], 0, '"a.dk b" dd'),
        ("a.dk\nb", ["ys", "e", ")"], 0, "(a.)dk\nb"),
        ("a.dk\nb", ["ys", "2e", "("], 0, "( a.dk )\nb"),
        ("abcd", ["$", "ys", "b", "]"], 0, "[abc]d"),
        ("ab.cd", ["$", "ys", "b", "'"], 3, "ab.'c'd"),
        ("ab.cd", ["$", "ys", "B", "{"], 0, "{ ab.c }d"),
        ("  abcd \n b", ["3l", "ys", "i", "w", '"'], 2, '  "abcd" \n b'),
        ("  abcd \n b", ["3l", "ys", "a", "w", '"'], 2, '  "abcd "\n b'),
        ("AB(CD)", ["l", "ys", "%", '"'], 1, 'A"B(CD)"'),
        ("AB(CD)", ["$", "ys", "%", '"'], 2, 'AB"(CD)"'),
        ("ABCD", ["ys", "f", "D", '"'], 0, '"ABCD"'),
        ("ABCD", ["c", "f", "F"], 0, "ABCD"),
        ("ABCD", ["ys", "t", "D", '"'], 0, '"ABC"D'),
        ("ABCD", ["$", "ys", "F", "A", '"'], 0, '"ABC"D'),
        ("ABCD", ["$", "ys", "T", "A", '"'], 1, 'A"BC"D'),
        ("AAAA", ["ys", "2f", "A", '"'], 0, '"AAA"A'),
        (' "AAA"', ["ys", "i", '"', ")"], 2, ' "(AAA)"'),
        (' "AAA"', ["ys", "a", '"', ")"], 0, '( "AAA")'),
        ("(AAA)", ["ys", "i", "(", '"'], 1, '("AAA")'),
        ("(AAA)", ["ys", "a", "(", '"'], 0, '"(AAA)"'),
        ("(AAA)", ["%", "ys", "a", ")", '"'], 0, '"(AAA)"'),
        ("(AAA)", ["ys", "s", "}"], 0, "{(AAA)}"),
    ],
)
def test_add_surround_in_normal(vim_bot, text, cmd_list, cursor_pos, text_expected):
    """Test to add surround in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected",
    [
        ("(a)", ["ds", ")"], 0, "a"),
        ("( a )", ["ds", "b"], 0, " a "),
        ("( a ) ( a )", ["ds", "("], 0, "a ( a )"),
        ("( a ) ( a )", ["ds", "(", "."], 2, "a a"),
        ("{a}", ["ds", "}"], 0, "a"),
        ("{ a }", ["ds", "B"], 0, " a "),
        ("{ a }", ["ds", "{"], 0, "a"),
        ("[ a ]", ["ds", "]"], 0, " a "),
        ("[ a ]", ["ds", "["], 0, "a"),
        ('"a" "a"', ["ds", '"'], 0, 'a "a"'),
        ('"a" "a"', ["ds", '"', "."], 2, "a a"),
        ("'a'", ["ds", "'"], 0, "a"),
        ("  (a)", ["ds", ")"], 2, "  a"),
        ("  'a'", ["ds", "'"], 2, "  a"),
        ("(a)  ", ["$", "ds", ")"], 4, "(a)  "),
        ("(a)  ", ["$", "ds", "'"], 4, "(a)  "),
        ("'a'  ", ["$", "ds", "'"], 4, "'a'  "),
        ("a  ", ["ds", ")"], 0, "a  "),
        ("a  ", ["ds", "'"], 0, "a  "),
    ],
)
def test_delete_surround_in_normal(vim_bot, text, cmd_list, cursor_pos, text_expected):
    """Test to delete surround in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected",
    [
        ("(a) (a)", ["cs", "b", "B"], 0, "{a} (a)"),
        ("(a) (a)", ["cs", "b", "B", "."], 4, "{a} {a}"),
        ("( a )", ["cs", "b", "'"], 0, "' a '"),
        ("( a )", ["cs", "(", "'"], 0, "'a'"),
        ("{ a }", ["cs", "B", '"'], 0, '" a "'),
        ("{ a }", ["cs", "{", "["], 0, "[ a ]"),
        ("[ a ]", ["cs", "[", "]"], 0, "[a]"),
        ("' a '", ["cs", "'", '"'], 0, '" a "'),
    ],
)
def test_change_surround_in_normal(vim_bot, text, cmd_list, cursor_pos, text_expected):
    """Test to change surround in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos



from __future__ import annotations

from bisect import bisect_left, bisect_right
from typing import Callable

from qtpy.QtCore import QPoint, QRegularExpression
from qtpy.QtGui import QTextCursor, QTextDocument
from qtpy.QtWidgets import QTextEdit
from spyder.config.manager import CONF

from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.motion import MotionInfo, MotionType


class SearchHelper:
    """Helper routines for search-related motions."""

    def __init__(self, vim_status, set_motion_info: Callable[..., MotionInfo]):
        self.vim_status = vim_status
        self.get_editor = vim_status.get_editor
        self.get_cursor = vim_status.get_cursor
        self._set_motion_info = set_motion_info

    # ------------------------------------------------------------------
    # Viewport helpers
    # ------------------------------------------------------------------
    def get_viewport_positions(self) -> tuple[int, int]:
        """Return start and end character positions of the visible viewport."""
        editor = self.get_editor()
        start_pos = editor.cursorForPosition(QPoint(0, 0)).position()
        bottom_right = QPoint(
            editor.viewport().width() - 1, editor.viewport().height() - 1
        )
        end_pos = editor.cursorForPosition(bottom_right).position()
        return start_pos, end_pos

    def search_forward_in_view(self, text: str) -> list[int]:
        """Return cursor positions of ``text`` inside the viewport."""
        editor = self.get_editor()
        cur_pos = editor.textCursor().position()
        view_start, view_end = self.get_viewport_positions()
        start_pos = min(view_end, cur_pos) + 1
        positions: list[int] = []
        while view_start <= start_pos <= view_end:
            cursor = editor.document().find(text, start_pos, QTextDocument.FindCaseSensitively)
            if cursor.isNull() or cursor.position() > view_end:
                break
            positions.append(cursor.position() - len(text))
            start_pos = cursor.position()
        return positions

    def search_backward_in_view(self, text: str) -> list[int]:
        """Return positions of ``text`` when searching backward in the viewport."""
        editor = self.get_editor()
        cur_pos = editor.textCursor().position()
        view_start, view_end = self.get_viewport_positions()
        start_pos = min(view_end, cur_pos)
        positions: list[int] = []
        while view_start <= start_pos <= view_end:
            cursor = editor.document().find(
                text,
                start_pos,
                QTextDocument.FindCaseSensitively | QTextDocument.FindBackward,
            )
            if cursor.isNull() or cursor.position() < view_start:
                break
            positions.append(cursor.position() - len(text))
            start_pos = cursor.position() - len(text) - 1
        return positions

    # ------------------------------------------------------------------
    # Sneak motions
    # ------------------------------------------------------------------
    def sneak(self, chars: str, num: int = 1, by_repeat_cmd: bool = False) -> MotionInfo:
        """Jump forward to the given two-character sequence within the viewport."""
        if not by_repeat_cmd:
            self.vim_status.find_info.set("s", chars)
        pos = None
        positions = self.search_forward_in_view(chars)
        if positions:
            pos = positions[(num - 1) % len(positions)]
        return self._set_motion_info(pos, motion_type=MotionType.CharWise)

    def display_another_group_after_sneak(self) -> None:
        """Show annotations for additional sneak targets."""
        positions = self.search_forward_in_view(self.vim_status.find_info.ch)
        info_group = {pos + 1: f"{idx};" if idx != 1 else ";" for idx, pos in enumerate(positions, 1)}
        self.vim_status.annotate_on_txt(info_group, timeout=1500)

    def rsneak(self, chars: str, num: int = 1, by_repeat_cmd: bool = False) -> MotionInfo:
        """Jump backward to the given two-character sequence within the viewport."""
        if not by_repeat_cmd:
            self.vim_status.find_info.set("S", chars)
        pos = None
        positions = self.search_backward_in_view(chars)
        if positions:
            pos = positions[(num - 1) % len(positions)]
        return self._set_motion_info(pos, motion_type=MotionType.CharWise)

    def display_another_group_after_rsneak(self) -> None:
        """Show annotations for additional reverse sneak targets."""
        positions = self.search_backward_in_view(self.vim_status.find_info.ch)
        info_group = {pos + 1: f"{idx};" if idx != 1 else ";" for idx, pos in enumerate(positions, 1)}
        self.vim_status.annotate_on_txt(info_group, timeout=1500)

    # ------------------------------------------------------------------
    # Search in document
    # ------------------------------------------------------------------
    def search(self, text: str) -> None:
        """Highlight all occurrences of ``text`` in the document."""
        editor = self.get_editor()
        cursor = QTextCursor(editor.document())
        cursor.movePosition(QTextCursor.Start)

        is_ignorecase = CONF.get(CONF_SECTION, "ignorecase")
        is_smartcase = CONF.get(CONF_SECTION, "smartcase")

        option = None
        if is_ignorecase:
            self.vim_status.search.ignorecase = True
            if is_smartcase and text.lower() != text:
                option = QTextDocument.FindCaseSensitively
                self.vim_status.search.ignorecase = False
        else:
            option = QTextDocument.FindCaseSensitively
            self.vim_status.search.ignorecase = False

        back = self.vim_status.search.color_bg
        fore = self.vim_status.search.color_fg
        search_stack = []
        while True:
            if option:
                cursor = editor.document().find(QRegularExpression(text), cursor, options=option)
            else:
                cursor = editor.document().find(QRegularExpression(text), cursor)

            if cursor.position() != -1:
                selection = QTextEdit.ExtraSelection()
                selection.format.setBackground(back)
                selection.format.setForeground(fore)
                selection.cursor = cursor
                search_stack.append(selection)
            else:
                break

        self.vim_status.cursor.set_extra_selections("vim_search", [i for i in search_stack])
        self.vim_status.search.selection_list = search_stack
        self.vim_status.search.txt_searched = text

    def n(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Move to the next search match."""
        positions = self.vim_status.search.get_sel_start_list()
        if not positions:
            return self._set_motion_info(None)
        text = self.vim_status.search.txt_searched
        self.vim_status.set_message(f"/{text}")
        cursor_pos = self.get_cursor().position()
        idx = bisect_right(positions, cursor_pos)
        if idx == len(positions):
            idx = 0
        idx = (idx + num - 1) % len(positions)
        return self._set_motion_info(positions[idx], motion_type=MotionType.CharWise)

    def N(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Move to the previous search match."""
        positions = self.vim_status.search.get_sel_start_list()
        if not positions:
            return self._set_motion_info(None)
        text = self.vim_status.search.txt_searched
        self.vim_status.set_message(f"?{text}")
        cursor_pos = self.get_cursor().position()
        idx = bisect_left(positions, cursor_pos)
        idx = (idx - (num - 1)) % len(positions)
        return self._set_motion_info(positions[idx - 1], motion_type=MotionType.CharWise)

    # ------------------------------------------------------------------
    # Word searches
    # ------------------------------------------------------------------
    def _get_word_under_cursor(self) -> str | None:
        editor = self.get_editor()
        return editor.get_current_word()

    def asterisk(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Search forward for the word under the cursor."""
        word = self._get_word_under_cursor()
        if word is None:
            return self._set_motion_info(None)
        self.search(fr"\b{word}\b")
        return self.n(num=num)

    def sharp(self, num: int = 1, num_str: str = "") -> MotionInfo:
        """Search backward for the word under the cursor."""
        word = self._get_word_under_cursor()
        if word is None:
            return self._set_motion_info(None)
        self.search(fr"\b{word}\b")
        return self.N(num=num)



# -*- coding: utf-8 -*-
"""Utilities for searching within the editor."""

from qtpy.QtCore import QRegularExpression
from qtpy.QtGui import (
    QBrush,
    QColor,
    QTextCursor,
    QValidator,
    QRegularExpressionValidator,
)

from spyder.config.manager import CONF

from spyder_okvim.spyder.config import CONF_SECTION


class SearchInfo:
    """Track search results inside the editor."""

    def __init__(self, vim_cursor):
        """Initialize the search state.

        Args:
            vim_cursor: Cursor helper used to draw selections.
        """
        self.color_fg = QBrush(QColor("#A9B7C6"))
        self.color_bg = QBrush(QColor("#30652F"))
        self.txt_searched = ""
        self.selection_list = []
        self.vim_cursor = vim_cursor
        self.ignorecase = True

        self.set_color()

    def set_color(self):
        """Refresh highlight colors from configuration."""
        self.color_fg = QBrush(QColor(CONF.get(CONF_SECTION, "search_fg_color")))
        self.color_bg = QBrush(QColor(CONF.get(CONF_SECTION, "search_bg_color")))

        for sel in self.selection_list:
            sel.format.setForeground(self.color_fg)
            sel.format.setBackground(self.color_bg)

    def get_sel_start_list(self):
        """Return start positions for valid selections."""
        cursor = self.vim_cursor.get_cursor()
        validator = QRegularExpressionValidator(
            QRegularExpression(self.txt_searched), None
        )

        tmp = []
        for sel in self.selection_list:
            cursor.setPosition(sel.cursor.selectionStart())
            cursor.setPosition(sel.cursor.selectionEnd(), QTextCursor.KeepAnchor)
            txt_sel = cursor.selectedText()

            if (validator.validate(txt_sel, 0)[0] == QValidator.Acceptable) or (
                self.ignorecase is True and txt_sel.lower() == self.txt_searched.lower()
            ):
                tmp.append(sel)
        self.selection_list = tmp

        self.vim_cursor.set_extra_selections(
            "vim_search", [i for i in self.selection_list]
        )

        return [i.cursor.selectionStart() for i in self.selection_list]





# -*- coding: utf-8 -*-
"""Tests for the executor_leader_key."""

# Standard Libraries
from unittest.mock import Mock

# Third Party Libraries
import pytest
from qtpy.QtCore import Qt
from spyder.api.plugins import Plugins


def test_auto_import(vim_bot):
    """Test auto_import."""
    _, _, editor, vim, qtbot = vim_bot

    editor.auto_import = Mock()
    editor.auto_import.auto_import = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "i")

    assert cmd_line.text() == ""
    assert editor.auto_import.auto_import.called


def test_toggle_breakpoint(vim_bot):
    """Test toggle breakpoint command."""
    _, editor_stack, editor, vim, qtbot = vim_bot

    editor.breakpoints_manager = Mock()
    editor.breakpoints_manager.toogle_breakpoint = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "b")

    assert cmd_line.text() == ""
    assert editor.breakpoints_manager.toogle_breakpoint.called


def test_run_cell_and_advance(vim_bot):
    """Test run_cell_and_advance."""
    _, editor_stack, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    cmd_line = vim.vim_cmd.commandline
    signal = editor_stack.sig_trigger_action
    expected_args = ("run cell and advance", Plugins.Run)
    with qtbot.waitSignal(signal, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyPress(cmd_line, Qt.Key_Enter)

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert tuple(blocker.args) == expected_args


def test_run_selection(vim_bot):
    """Test run_selection."""
    _, editor_stack, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    cmd_line = vim.vim_cmd.commandline
    signal = editor_stack.sig_trigger_action
    expected_args = ("run selection and advance", Plugins.Run)
    with qtbot.waitSignal(signal, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyClicks(cmd_line, "r")

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert tuple(blocker.args) == expected_args

    qtbot.keyClicks(cmd_line, "Vj")
    with qtbot.waitSignal(signal, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyClicks(cmd_line, "r")

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert tuple(blocker.args) == expected_args


def test_formatting(vim_bot):
    """Test formatting."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    editor.format_document_or_range = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "f")

    assert cmd_line.text() == ""
    assert editor.format_document_or_range.called

    editor.format_document_or_range = Mock()
    qtbot.keyClicks(cmd_line, "ggVj")

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "f")

    assert cmd_line.text() == ""
    assert editor.format_document_or_range.called


def test_open_switcher_plugin(vim_bot):
    """Test open switcher when provided by plugin."""
    main, _, _, vim, qtbot = vim_bot

    if hasattr(main, "open_switcher"):
        delattr(main, "open_switcher")

    plugin = Mock()
    main.get_plugin = Mock(return_value=plugin)

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "p")

    assert cmd_line.text() == ""
    assert plugin.open_switcher.called


def test_open_symbol_swithcer_plugin(vim_bot):
    """Test open symbol switcher when provided by plugin."""
    main, _, _, vim, qtbot = vim_bot

    if hasattr(main, "open_switcher"):
        delattr(main, "open_switcher")

    plugin = Mock()
    main.get_plugin = Mock(return_value=plugin)

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "s")

    assert cmd_line.text() == ""
    assert plugin.open_switcher.called



# -*- coding: utf-8 -*-
"""Executor for normal mode commands.

This module implements :class:`ExecutorNormalCmd`, which interprets keystrokes
when the editor is in *normal* mode.  Commands are dispatched to helper classes
for motion and editing actions.  The executor can also switch into various
submodes such as search or register selection.
"""

# Standard Libraries
import re

# Third Party Libraries
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QKeyEvent, QTextCursor
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.executor.decorators import submode
from spyder_okvim.executor.executor_base import (
    FUNC_INFO,
    RETURN_EXECUTOR_METHOD_INFO,
    ExecutorBase,
)
from spyder_okvim.executor.executor_colon import ExecutorColon
from spyder_okvim.executor.executor_easymotion import ExecutorEasymotion
from spyder_okvim.executor.executor_sub import (
    ExecutorSearch,
    ExecutorSubCmd_alnum,
    ExecutorSubCmd_closesquarebracket,
    ExecutorSubCmd_closebrace,
    ExecutorSubCmd_f_t,
    ExecutorSubCmd_g,
    ExecutorSubCmd_opensquarebracket,
    ExecutorSubCmd_openbrace,
    ExecutorSubCmd_r,
    ExecutorSubCmd_register,
    ExecutorSubCmd_z,
    ExecutorSubCmd_Z,
    ExecutorSubCmdSneak,
    ExecutorSubMotion,
    ExecutorSubMotion_c,
    ExecutorSubMotion_d,
    ExecutorSubMotion_y,
)
from spyder_okvim.executor.mixins import MovementMixin
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.motion import MotionInfo, MotionType


class ExecutorNormalCmd(MovementMixin, ExecutorBase):
    """Executor for normal mode."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        # MovementMixin hooks
        self.move_cursor = self.vim_status.cursor.set_cursor_pos
        self.move_cursor_no_end = self.vim_status.cursor.set_cursor_pos_without_end

        cmds = (
            "aAiIvVhHjpPyJkKlLMoOruwWbBegGsSxdcDCnN^$~:%fFtT\"`'m;,.zZ/<>{} \b\rq@\[\]*#"
        )
        cmds = "".join(re.escape(c) for c in cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.apply_motion_info_in_normal = (
            self.vim_status.cursor.apply_motion_info_in_normal
        )
        self.apply_motion_info_in_yank = (
            self.vim_status.cursor.apply_motion_info_in_yank
        )
        self.executor_colon = ExecutorColon(vim_status)

        self.executor_sub_g = ExecutorSubCmd_g(vim_status)
        self.executor_sub_f_t = ExecutorSubCmd_f_t(vim_status)
        self.executor_sub_r = ExecutorSubCmd_r(vim_status)
        self.executor_sub_z = ExecutorSubCmd_z(vim_status)
        self.executor_sub_Z = ExecutorSubCmd_Z(vim_status)
        self.executor_sub_motion = ExecutorSubMotion(vim_status)
        self.executor_sub_motion_c = ExecutorSubMotion_c(vim_status)
        self.executor_sub_motion_d = ExecutorSubMotion_d(vim_status)
        self.executor_sub_motion_y = ExecutorSubMotion_y(vim_status)
        self.executor_sub_register = ExecutorSubCmd_register(vim_status)
        self.executor_sub_search = ExecutorSearch(vim_status)
        self.executor_sub_alnum = ExecutorSubCmd_alnum(vim_status)
        self.executor_sub_easymotion = ExecutorEasymotion(vim_status)
        self.executor_sub_sneak = ExecutorSubCmdSneak(vim_status)
        self.executor_sub_opensquarebracekt = ExecutorSubCmd_opensquarebracket(
            vim_status
        )
        self.executor_sub_closesquarebracekt = ExecutorSubCmd_closesquarebracket(
            vim_status
        )
        self.executor_sub_openbrace = ExecutorSubCmd_openbrace(vim_status)
        self.executor_sub_closebrace = ExecutorSubCmd_closebrace(vim_status)

    def a(self, num=1, num_str=""):
        """Append text after the cursor."""
        self.vim_status.update_dot_cmd(connect_editor=True)

        cursor = self.get_cursor()
        if not (cursor.atBlockEnd() and cursor.atBlockStart()):
            cursor.movePosition(QTextCursor.Right)
            self.set_cursor(cursor)
        self.get_editor().setFocus()

    def A(self, num=1, num_str=""):
        """Append text at the end of the line."""
        self.vim_status.update_dot_cmd(connect_editor=True)

        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.EndOfLine)
        self.set_cursor(cursor)
        self.get_editor().setFocus()

    def i(self, num=1, num_str=""):
        """Insert text before the cursor."""
        self.vim_status.update_dot_cmd(connect_editor=True)
        self.get_editor().setFocus()

    def I(self, num=1, num_str=""):
        """Insert text before the first non-blank in the line."""
        self.vim_status.update_dot_cmd(connect_editor=True)

        motion_info = self.helper_motion.caret(num=num)

        cursor = self.get_cursor()
        cursor.setPosition(motion_info.cursor_pos)
        self.set_cursor(cursor)

        self.get_editor().setFocus()

    def v(self, num=1, num_str=""):
        """Start visual mode per character."""
        self.vim_status.to_visual_char()

    def V(self, num=1, num_str=""):
        """Start visual mode per line."""
        self.vim_status.to_visual_line()

    def J(self, num=1, num_str=""):
        """Join lines, with a minimum of two lines."""
        cursor = self.get_cursor()

        if num > 1:
            num -= 1

        block_no_start = cursor.blockNumber()
        block_no_end = cursor.blockNumber() + num
        cursor_pos_start = cursor.position()

        self.helper_action.join_lines(cursor_pos_start, block_no_start, block_no_end)

    def o(self, num=1, num_str=""):
        """Begin a new line below the cursor and insert text."""
        editor = self.get_editor()
        cursor = editor.textCursor()
        cursor.movePosition(QTextCursor.EndOfLine)
        editor.setTextCursor(cursor)
        editor.setFocus()

        # Send the keyevent to editor for using autoindent of editor.
        new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Return, Qt.NoModifier)
        editor.keyPressEvent(new_event)

        self.vim_status.update_dot_cmd(connect_editor=True)

    def O(self, num=1, num_str=""):
        """Begin a new line above the cursor and insert text."""
        editor = self.get_editor()
        cursor = editor.textCursor()
        cursor.movePosition(QTextCursor.StartOfLine)
        cursor.insertText("\n")
        cursor.movePosition(QTextCursor.Up)
        editor.setTextCursor(cursor)
        editor.setFocus()

        self.vim_status.update_dot_cmd(connect_editor=True)

    def K(self, num=1, num_str=""):
        """Show the help for the keyword under the cursor."""
        self.get_editorstack().inspect_current_object()
        # self.vim_status.set_focus_to_vim_after_delay(500)

    def u(self, num=1, num_str=""):
        """Undo changes."""
        editor = self.get_editor()
        n_block_old = editor.blockCount()

        for _ in range(num):
            editor.undo()
        cursor = editor.textCursor()
        pos = cursor.position()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            cursor.movePosition(QTextCursor.Left)
            pos = cursor.position()

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            if n_block_new > n_block_old:
                self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")
            else:
                self.vim_status.set_message(f"{n_block_old - n_block_new} fewer lines")
        else:
            self.vim_status.set_message(f"{num} changes")

        self.set_cursor_pos(pos)

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def g(self, num=1, num_str=""):
        """Start g submode."""
        return self.executor_sub_g

    def G(self, num=1, num_str=""):
        """Move to the line."""
        self.vim_status.push_jump()
        motion_info = self.helper_motion.G(num=num, num_str=num_str)

        self.set_cursor_pos(motion_info.cursor_pos)
        self.vim_status.push_jump()

    def tilde(self, num=1, num_str=""):
        """Switch case of the character under the cursor.

        Move the cursor to the right.
        """
        motion_info = self.helper_motion.l(num, num_str=num_str)

        self.helper_action.handle_case(motion_info, "swap")
        self.set_cursor_pos(motion_info.cursor_pos)

        cursor = self.get_cursor()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            self.h(1)

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def f(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        return self.executor_sub_f_t

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def F(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        return self.executor_sub_f_t

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def t(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        return self.executor_sub_f_t

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def T(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        return self.executor_sub_f_t

    def r(self, num=1, num_str=""):
        """Replace the ch under the cursor with input."""
        cursor = self.get_cursor()
        pos_start = cursor.position()
        cursor.movePosition(QTextCursor.EndOfBlock)
        pos_block_end = cursor.position()
        if pos_block_end - pos_start < num:
            pos_end = pos_start
        else:
            pos_end = pos_start + num

        executor_sub = self.executor_sub_r
        executor_sub.pos_start = pos_start
        executor_sub.pos_end = pos_end

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def dot(self, num=1, num_str=""):
        """Run previous change."""
        # avoid to execute . command twice.
        self.vim_status.manager_macro.remove_last_key(".")
        cmd_str = self.vim_status.dot_cmd.to_cmd_string(num, num_str)

        if not cmd_str:
            return

        cmd_line = self.vim_status.cmd_line
        self.vim_status.running_dot_cmd = True
        cmd_line.clear()
        for ch in cmd_str:
            cmd_line.setText(cmd_line.text() + ch)

        for key_info in self.vim_status.dot_cmd.key_list_to_cmd_line:
            event = key_info.to_event()
            cmd_line.keyPressEvent(event)

        editor = self.get_editor()
        for key_info in self.vim_status.dot_cmd.key_list_from_editor:
            event = key_info.to_event()
            editor.keyPressEvent(event)

        self.vim_status.running_dot_cmd = False

        self.vim_status.set_focus_to_vim()

        cursor_pos = self.get_cursor().position()
        self.vim_status.cursor.set_cursor_pos_without_end(cursor_pos)

    def z(self, num=1, num_str=""):
        """Start z submode."""
        executor_sub = self.executor_sub_z

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred([])

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def Z(self, num=1, num_str=""):
        """Start Z submode."""
        executor_sub = self.executor_sub_Z

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred([])

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def greater(self, num=1, num_str=""):
        """Shift lines rightwards."""
        executor_sub = self.executor_sub_motion

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.helper_action.indent, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def less(self, num=1, num_str=""):
        """Shift lines leftwards."""
        executor_sub = self.executor_sub_motion

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.helper_action.unindent, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def quote(self, num=1, num_str=""):
        """Set the name of register."""
        executor_sub = self.executor_sub_register

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred([])

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def p(self, num=1, num_str=""):
        """Put the text from register after the cursor."""
        self.helper_action.paste_in_normal(num, is_lower=True)

    def P(self, num=1, num_str=""):
        """Put the text from register after the cursor."""
        self.helper_action.paste_in_normal(num, is_lower=False)

    def y(self, num=1, num_str=""):
        """Yank {motion} text into register."""
        executor_sub = self.executor_sub_motion_y

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(lambda x: self.helper_action.yank(x, is_explicit=True), True),
                FUNC_INFO(self.apply_motion_info_in_yank, True),
            ]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def s(self, num=1, num_str=""):
        """Delete characters and start insert."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak

            self.set_parent_info_to_submode(executor_sub, num, num_str)

            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.apply_motion_info_in_normal, True),
                    FUNC_INFO(
                        self.helper_motion.display_another_group_after_sneak, False
                    ),
                ]
            )
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            motion_info = self.helper_motion.l(num, num_str)
            self.helper_action.yank(motion_info)
            self.helper_action.delete(motion_info, is_insert=True)
            self.get_editor().setFocus()

    def S(self, num=1, num_str=""):
        """Delete characters and start insert."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak

            self.set_parent_info_to_submode(executor_sub, num, num_str)

            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.apply_motion_info_in_normal, True),
                    FUNC_INFO(
                        self.helper_motion.display_another_group_after_rsneak, False
                    ),
                ]
            )
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            cursor = self.get_cursor()
            block = cursor.block()
            txt = block.text()
            n_space = len(txt) - len(txt.lstrip())

            motion_info = self.helper_motion.j(num - 1, num_str)
            self.helper_action.yank(motion_info)
            self.helper_action.delete(
                motion_info, is_insert=True, replace_txt=" " * n_space
            )
            self.get_editor().setFocus()

    def x(self, num=1, num_str=""):
        """Delete characters and start insert."""
        cursor_pos = self.get_cursor().position()

        motion_info = self.helper_motion.l(num, num_str)
        self.helper_action.yank(motion_info)
        self.helper_action.delete(motion_info, is_insert=False)

        self.vim_status.cursor.set_cursor_pos_without_end(cursor_pos)

    def D(self, num=1, num_str=""):
        """Delete the characters under the cursor until the end of the line.

        If count is greater than 1, This command deletes [count]-1 more lines.
        (not linewise)
        """
        editor = self.get_editor()
        n_block = editor.blockCount()
        block_no_start = self.get_cursor().blockNumber()
        block_no_end = min([n_block - 1, block_no_start + num - 1])
        block = editor.document().findBlockByNumber(block_no_end)

        motion_info = MotionInfo()
        motion_info.cursor_pos = block.position() + block.length() - 1
        motion_info.motion_type = MotionType.CharWise
        sel_start, sel_end = self.helper_action.yank(motion_info)
        self.helper_action.delete(motion_info, is_insert=False)

        self.vim_status.cursor.set_cursor_pos_without_end(sel_start)

    def C(self, num=1, num_str=""):
        """Execute D command and start insert mode."""
        editor = self.get_editor()
        n_block = editor.blockCount()
        block_no_start = self.get_cursor().blockNumber()
        block_no_end = min([n_block - 1, block_no_start + num - 1])
        block = editor.document().findBlockByNumber(block_no_end)

        motion_info = MotionInfo()
        motion_info.cursor_pos = block.position() + block.length() - 1
        motion_info.motion_type = MotionType.CharWise
        sel_start, sel_end = self.helper_action.yank(motion_info)
        self.helper_action.delete(motion_info, is_insert=True)

        self.vim_status.cursor.set_cursor_pos(sel_start)
        self.get_editor().setFocus()

    def d(self, num=1, num_str=""):
        """Delete text."""
        executor_sub = self.executor_sub_motion_d

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(self.helper_action.yank, True),
                FUNC_INFO(
                    lambda x: self.helper_action.delete(x, is_insert=False), True
                ),
            ]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def c(self, num=1, num_str=""):
        """Delete text and start insert."""
        executor_sub = self.executor_sub_motion_c

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(self.helper_action.yank, True),
                FUNC_INFO(lambda x: self.helper_action.delete(x, is_insert=True), True),
                FUNC_INFO(self.get_editor().setFocus, False),
            ]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def slash(self, num=1, num_str=""):
        """Go to the next searched text."""
        self.vim_status.set_message("")
        executor_sub = self.executor_sub_search

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_normal, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, False)

    def n(self, num=1, num_str=""):
        """Go to the next searched text."""
        self.vim_status.push_jump()
        motion_info = self.helper_motion.n(num=num, num_str=num_str)

        self.set_cursor_pos(motion_info.cursor_pos)
        self.vim_status.push_jump()

    def N(self, num=1, num_str=""):
        """Go to the previous searched text."""
        self.vim_status.push_jump()
        motion_info = self.helper_motion.N(num=num, num_str=num_str)

        self.set_cursor_pos(motion_info.cursor_pos)
        self.vim_status.push_jump()

    def m(self, num=1, num_str=""):
        """Set bookmark with given name."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.vim_status.set_bookmark, True)]
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def apostrophe(self, num=1, num_str=""):
        """Jump to bookmark linewise."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch):
            self.vim_status.push_jump()
            if ch.isupper():
                self.vim_status.jump_to_bookmark(ch)
                cursor = self.get_cursor()
                self.vim_status.cursor.set_cursor_pos(cursor.block().position())
            else:
                info = self.helper_motion.apostrophe(ch)
                if info.cursor_pos is not None:
                    self.vim_status.cursor.set_cursor_pos(info.cursor_pos)
            self.vim_status.push_jump()

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def backtick(self, num=1, num_str=""):
        """Jump to bookmark charwise."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch):
            self.vim_status.push_jump()
            if ch.isupper():
                self.vim_status.jump_to_bookmark(ch)
            else:
                info = self.helper_motion.backtick(ch)
                if info.cursor_pos is not None:
                    self.vim_status.cursor.set_cursor_pos(info.cursor_pos)
            self.vim_status.push_jump()

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def q(self, num=1, num_str=""):
        """Record typed characters into register."""
        if self.vim_status.is_recording_macro():
            self.vim_status.stop_recording_macro()
            self.vim_status.set_message("")
        else:
            executor_sub = self.executor_sub_alnum

            self.set_parent_info_to_submode(executor_sub, num, num_str)

            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.vim_status.start_recording_macro, True),
                    FUNC_INFO(lambda: self.vim_status.set_message(""), False),
                ]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def at(self, num=1, num_str=""):
        """Execute contents of register."""
        if self.vim_status.is_recording_macro():
            self.vim_status.manager_macro.remove_last_key("@")
            return

        executor_sub = self.executor_sub_alnum

        manager_macro = self.vim_status.manager_macro
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(lambda x: manager_macro.set_info_for_execute(x, num), True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def run_easymotion(self, num=1, num_str=""):
        """Run easymotion."""
        executor_sub = self.executor_sub_easymotion

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_normal, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def opensquarebracket(self, num=1, num_str=""):
        """Start [ submode."""
        executor_sub = self.executor_sub_opensquarebracekt

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        return executor_sub

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def closesquarebracket(self, num=1, num_str=""):
        """Start ] submode."""
        executor_sub = self.executor_sub_closesquarebracekt

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        return executor_sub

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def openbrace(self, num=1, num_str=""):
        """Start { submode."""
        executor_sub = self.executor_sub_openbrace

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        return executor_sub

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_normal, True)])
    def closebrace(self, num=1, num_str=""):
        """Start } submode."""
        executor_sub = self.executor_sub_closebrace

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        return executor_sub

    def asterisk(self, num=1, num_str=""):
        """Search word under cursor forward."""
        motion_info = self.helper_motion.asterisk(num)
        self.apply_motion_info_in_normal(motion_info)

    def sharp(self, num=1, num_str=""):
        """Search word under cursor backward."""
        motion_info = self.helper_motion.sharp(num)
        self.apply_motion_info_in_normal(motion_info)





# -*- coding: utf-8 -*-
"""Tests for :mod:`spyder_okvim.executor.executor_colon`."""

# Third Party Libraries
import pytest
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QKeyEvent


@pytest.mark.parametrize(
    "text, cmd_list, cmd_line_expected",
    [
        ("", [":", "k", "k"], ":kk"),
        ("", [":", "k", "k", Qt.Key_Escape], ""),
    ],
)
def test_colon_cmd(vim_bot, text, cmd_list, cmd_line_expected):
    """Test colon command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == cmd_line_expected


@pytest.mark.parametrize(
    "text, cmd_list",
    [
        ("", [":", Qt.Key_Return]),
        ("", [":", Qt.Key_Left, "d", Qt.Key_Enter]),
    ],
)
def test_colon_corner_case_cmd(vim_bot, text, cmd_list):
    """Test colon command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.sub_mode is None


def test_colon_w_command(vim_bot):
    """Test :w."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "w")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.save.assert_called_once_with()
    main.editor.save.reset_mock()


def test_colon_q_command(vim_bot):
    """Test :q."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.close_file.assert_called_once_with()
    main.editor.close_file.reset_mock()


def test_colon_q_command_without_close_file(vim_bot):
    """Test :q when close_file API is missing."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    orig_close_file = main.editor.close_file
    delattr(main.editor, "close_file")
    with pytest.raises(AttributeError):
        vim.vim_cmd.executor_normal_cmd.executor_colon.q()
    main.editor.close_file = orig_close_file


def test_colon_qexclamation_command(vim_bot):
    """Test :q!."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyClicks(cmd_line, "!")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.close_file.assert_called_once_with()
    main.editor.close_file.reset_mock()


def test_colon_wq_command(vim_bot):
    """Test :wq."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "w")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.save.assert_called_once_with()
    main.editor.close_file.assert_called_once_with()
    main.editor.save.reset_mock()
    main.editor.close_file.reset_mock()


def test_colon_n_command(vim_bot):
    """Test :n."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "n")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.new.assert_called_once_with()


def test_colon_backspace_command(vim_bot):
    """Test backspace in ex cmd."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "n")
    qtbot.keyPress(cmd_line, Qt.Key_Backspace)

    assert cmd_line.text() == ":"
    assert vim.vim_cmd.vim_status.sub_mode is not None

    qtbot.keyPress(cmd_line, Qt.Key_Backspace)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.sub_mode is None


def test_colon_marks_command(vim_bot, monkeypatch):
    """Test :marks opens dialog and jumps to the selected mark."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vs = vim.vim_cmd.vim_status
    vs.cursor.set_cursor_pos(2)
    vs.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    vs.cursor.set_cursor_pos(0)

    # Project Libraries
    from spyder_okvim.utils import mark_dialog

    monkeypatch.setattr(mark_dialog.MarkListDialog, "exec_", lambda self: None)
    monkeypatch.setattr(
        mark_dialog.MarkListDialog, "get_selected_mark", lambda self: "a"
    )

    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "marks")
    qtbot.keyPress(cmd_line, Qt.Key_Return)

    assert editor.textCursor().position() == 2


def test_colon_marks_escape_focus(vim_bot, monkeypatch):
    """Focus returns to command line when dialog closed with ESC."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vs = vim.vim_cmd.vim_status
    pos_before = editor.textCursor().position()

    # Project Libraries
    from spyder_okvim.utils import mark_dialog

    monkeypatch.setattr(mark_dialog.MarkListDialog, "exec_", lambda self: 0)
    # Return a mark even though the dialog was rejected to ensure the
    # executor ignores it when ESC is pressed.
    monkeypatch.setattr(
        mark_dialog.MarkListDialog, "get_selected_mark", lambda self: "a"
    )

    called = []
    monkeypatch.setattr(vs, "jump_to_bookmark", lambda *a: called.append(True))

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "marks")
    qtbot.keyPress(cmd_line, Qt.Key_Return)

    assert editor.textCursor().position() == pos_before
    assert cmd_line.hasFocus()
    assert not called


def test_colon_jumps_dialog_navigation(vim_bot, monkeypatch):
    """Jump dialog tracks jumplist navigation across files."""
    _, stack, editor, vim, qtbot = vim_bot
    vs = vim.vim_cmd.vim_status
    orig_file = stack.get_current_filename()
    stack.set_current_filename(stack.get_filenames()[0])
    vs.cursor.set_cursor_pos(0)
    vs.reset_for_test()

    # Create jumps across multiple files
    vs.push_jump()
    stack.set_current_filename(stack.get_filenames()[1])
    vs.push_jump()
    stack.set_current_filename(stack.get_filenames()[2])
    vs.push_jump()

    # Project Libraries
    from spyder_okvim.utils import jump_dialog

    captured = {}

    def fake_exec(self):
        captured["dlg"] = self
        return None

    monkeypatch.setattr(jump_dialog.JumpListDialog, "exec_", fake_exec)

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":jumps")
    qtbot.keyPress(cmd_line, Qt.Key_Return)

    dlg = captured["dlg"]
    dlg.list_viewer.setFocus()
    assert dlg.list_model.rowCount() == 3
    assert dlg.list_viewer.currentIndex().row() == vs.jump_list.index - 1

    qtbot.keyPress(dlg.list_viewer, Qt.Key_O, modifier=Qt.ControlModifier)
    assert stack.get_current_filename() == stack.get_filenames()[1]
    assert dlg.list_viewer.currentIndex().row() == vs.jump_list.index - 1

    qtbot.keyPress(dlg.list_viewer, Qt.Key_I, modifier=Qt.ControlModifier)
    assert stack.get_current_filename() == stack.get_filenames()[2]
    assert dlg.list_viewer.currentIndex().row() == vs.jump_list.index - 1

    dlg.reject()
    assert cmd_line.hasFocus()

    # Restore state for other tests
    stack.set_current_filename(orig_file)



# -*- coding: utf-8 -*-
"""Utilities to inspect Spyder code cells."""

from __future__ import annotations

# Standard Libraries
from dataclasses import dataclass
from typing import TYPE_CHECKING

# Third Party Libraries
from qtpy.QtGui import QTextDocument

if TYPE_CHECKING:  # pragma: no cover - imports for typing only
    from spyder.plugins.editor.widgets.codeeditor.codeeditor import CodeEditor


@dataclass(frozen=True)
class CellRegion:
    """Describe a Spyder code cell within a document."""

    index: int
    name: str | None
    level: int
    header: str | None
    block_number: int
    start_line: int
    end_line: int
    start_position: int
    end_position: int

    @property
    def has_header(self) -> bool:
        """Return ``True`` when this region originates from a cell header."""
        return self.header is not None


def _document_end_position(document: QTextDocument) -> int:
    """Return the exclusive document end position."""
    last_block = document.lastBlock()
    return last_block.position() + last_block.length()


def _safe_block(document: QTextDocument, block_number: int):
    """Return a valid text block for ``block_number``."""
    block = document.findBlockByNumber(block_number)
    if block.isValid():
        return block
    return document.lastBlock()


def _fallback_region(document: QTextDocument) -> list[CellRegion]:
    """Return a single region covering the entire document."""
    end_line = document.lastBlock().blockNumber()
    return [
        CellRegion(
            index=0,
            name=None,
            level=0,
            header=None,
            block_number=0,
            start_line=0,
            end_line=end_line,
            start_position=0,
            end_position=_document_end_position(document),
        )
    ]


def get_document_cells(editor: CodeEditor | None) -> list[CellRegion]:
    """Return all cell regions defined in ``editor``.

    Args:
        editor: The active Spyder code editor instance.

    Returns:
        A list of :class:`CellRegion` objects ordered by their appearance in
        the document. When no explicit cell headers are present the whole file
        is returned as a single region.
    """
    if editor is None:
        return []

    document = editor.document()
    cell_entries = editor.get_cell_list()
    if not cell_entries:
        return _fallback_region(document)

    regions: list[CellRegion] = []
    total_blocks = document.blockCount()

    for idx, (block_number, oedata) in enumerate(cell_entries):
        start_block = _safe_block(document, block_number)
        start_line = start_block.blockNumber()
        start_pos = start_block.position()

        if idx + 1 < len(cell_entries):
            next_block_number = cell_entries[idx + 1][0]
            end_line = min(total_blocks - 1, max(block_number, next_block_number - 1))
            end_pos = _safe_block(document, next_block_number).position()
        else:
            end_line = total_blocks - 1
            end_pos = _document_end_position(document)

        name = getattr(oedata, "def_name", None) or None
        header = getattr(oedata, "text", None) or None
        level = getattr(oedata, "cell_level", 0)

        regions.append(
            CellRegion(
                index=idx,
                name=name,
                level=level,
                header=header,
                block_number=block_number,
                start_line=start_line,
                end_line=end_line,
                start_position=start_pos,
                end_position=end_pos,
            )
        )

    return regions




"""Executors for manipulating surrounding characters."""

# Project Libraries
from spyder_okvim.executor.executor_base import ExecutorSubBase
from spyder_okvim.utils.motion import MotionInfo, MotionType

SURROUNDINGS = "'\"{[()]}"


class ExecutorAddSurround(ExecutorSubBase):
    """Executor for adding surrounding."""

    def __init__(self, vim_status):
        """Initialize submode for adding surroundings."""
        super().__init__(vim_status)
        self.allow_leaderkey = False
        self.has_zero_cmd = False

        self.pos_start = 0
        self.pos_end = 0

    def set_motion_info(self, motion_info: MotionInfo):
        """Set motion_info."""
        if motion_info.motion_type == MotionType.BlockWise:
            sel_start = motion_info.sel_start
            sel_end = motion_info.sel_end
        else:
            cursor = self.get_cursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            sel_start, sel_end = sorted([cursor_pos_cur, cursor_pos_new])

            if motion_info.motion_type == MotionType.CharWiseIncludingEnd:
                sel_end += 1
            elif motion_info.motion_type == MotionType.LineWise:
                block_start, _ = self.vim_status.cursor.get_block(sel_start)
                block_end, _ = self.vim_status.cursor.get_block(sel_end)
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1

        self.pos_start = sel_start
        self.pos_end = sel_end

    def __call__(self, ch: str):
        """Add ``ch`` around the motion range."""
        self.vim_status.sub_mode = None

        ch = ch.replace("b", ")")
        ch = ch.replace("B", "}")

        if ch in SURROUNDINGS:
            self.update_input_cmd_info(None, None, ch)
            self.helper_action.add_surrounding(self.pos_start, self.pos_end, ch)

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(self.pos_start)

        return True


class ExecutorDeleteSurround(ExecutorSubBase):
    """Executor for deleting surrounding."""

    def __init__(self, vim_status):
        """Initialize submode for deleting surroundings."""
        super().__init__(vim_status)
        self.allow_leaderkey = False
        self.has_zero_cmd = False

    def __call__(self, ch: str):
        """Delete the surrounding character ``ch``."""
        self.vim_status.sub_mode = None

        ch = ch.replace("b", ")")
        ch = ch.replace("B", "}")

        if ch in SURROUNDINGS:
            self.update_input_cmd_info(None, None, ch)
            motion_info = self.helper_action.delete_surrounding(ch)
            if isinstance(motion_info.sel_start, int):
                self.vim_status.to_normal()
                self.vim_status.cursor.set_cursor_pos(motion_info.sel_start)

        return True


class ExecutorChangeSurround(ExecutorSubBase):
    """Executor for chaging surrounding."""

    def __init__(self, vim_status):
        """Initialize submode for changing surroundings."""
        super().__init__(vim_status)
        self.allow_leaderkey = False
        self.has_zero_cmd = False

    def __call__(self, txt: str):
        """Replace the existing surrounding with ``txt``."""
        if len(txt) < 2:
            return False

        self.vim_status.sub_mode = None

        txt = txt.replace("b", ")")
        txt = txt.replace("B", "}")

        if txt[0] in SURROUNDINGS and txt[1] in SURROUNDINGS:
            self.update_input_cmd_info(None, None, txt)
            motion_info = self.helper_action.change_surrounding(txt[0], txt[1])
            if isinstance(motion_info.sel_start, int):
                self.vim_status.to_normal()
                self.vim_status.cursor.set_cursor_pos(motion_info.sel_start)

        return True
